/*****************************************************************************/
/*
                              CGIplusSkel.c

A skeletal program serving as a template for CGI/CGIplus-capable scripts.

Uses a sophisticated single function for accessing CGI variables. The function
differentiates between standard and CGIplus environments and changes behaviour
accordingly.  The function also serves to block waiting for subsequent CGIplus
requests.


08-JUN-97  MGD  v1.0.0, initial development
*/
/*****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

int  IsCgiPlus = 0;
char  *CgiPlusEofPtr = NULL;

/* required prototypes */
char* CgiVar (char*);

/*****************************************************************************/
/*
*/

main ()

{
   int  Count = 0;
   char  *RemoteHostPtr;

   IsCgiPlus = ((CgiPlusEofPtr = getenv("CGIPLUSEOF")) != NULL);

   do {

      /* with CGIplus this call will block waiting for the next request */
      CgiVar ("");

      RemoteHostPtr = CgiVar ("WWW_REMOTE_HOST");

      fprintf (stdout,
"Content-Type: text/html\n\
Expires: Thu, 01 Jan 1970 00:00:01 GMT\n\
\n\
%d. <B>Hello world</B> ... well the \
<FONT SIZE=+1>&nbsp;<TT>%s</TT>&nbsp;</FONT> part of it anyway!\n",
               ++Count, RemoteHostPtr);

      /* record-oriented <stdout>, no need to flush, write end-of-output line */
      if (IsCgiPlus) fprintf (stdout, "%s\n", CgiPlusEofPtr);

   } while (IsCgiPlus);

   exit (1);
}

/*****************************************************************************/
/*
This function provides a simple, uniform mechanism for a C Language script to
access CGI variable values transparently in both the standard CGI and the
CGIplus environments.  Requires the global storage 'CgiPlusEofPtr' to be
non-NULL to indicate a CGIplus environment.

It is completely self-contained. Simply call the function with the CGI
variable name value to be accessed and if that CGI variable exists a pointer
to the value will be returned.  Errors and non-existant variables return NONE
(see below).  The values pointed to should NOT be modified in any way (copy the
value if this is required). 

Supplying an empty variable name (i.e. "") will cause the function to block
until the arrival of a request, upon which it just returns a NONE pointer,
allowing start-of-request synchronization.  This will also release allocated
memory and reset in preparation for reading the next request, an alternative to
calling with a NULL parameter at the end of request processing (see below).

For the CGIplus environment supplying a wilcard variable name (i.e. "*")
returns each CGIplus variable string (line from CGIPLUSIN) in turn, with no
more being indicated by a NONE pointer.

At the end of a CGIplus request call the function with a NULL parameter to
release allocated memory and reset ready for processing the next request's
variables.  It is not necessary to do this if calling with an empty parameter
to synchronize on the start of a request (see above).

The CGIplus functionality works by initially reading all lines from
CGIPLUSIN until the terminating blank line is encountered. These lines are
stored as a series of contiguous null-terminated strings, terminated by an
empty string, in allocated memory. These strings are searched for the required
variable name and a pointer to its value returned if found.

The value NONE shold be set according to the requirements of the application
the function is included within.  If calling functions can handle NULL pointers
being returned then set the value to NULL.  If non-NULL pointers are prefered
then set the value to an empty string.
*/

char* CgiVar (char *VarName)

{
/** #  define NONE NULL **/
#  define NONE ""
#  define CGIPLUS_LINE_SIZE 1024

   static int  VarPtrSize = 0;
   static char  *NextPtr = NULL,
                *VarPtr = NULL;

   register char  *cptr, *sptr;

   /*********/
   /* begin */
   /*********/

   if (VarName == NULL || !VarName[0])
   {
      /* end of request, release memory and reinitialize */
      if (VarPtr != NULL)
      {
         free (VarPtr);
         NextPtr = VarPtr = NULL;
      }
      if (VarName == NULL) return (NONE);
   }

   if (CgiPlusEofPtr == NULL)
   {
      /* standard CGI environment, get variable from environment */
      if (!VarName[0]) return (NONE);
      if ((cptr = getenv(VarName)) == NULL)
         return (NONE);
      else
         return (cptr);
   }

   if (VarPtr == NULL)
   {
      /* read lines containing CGIplus variables from <CGIPLUSIN> */
      int  LineLength;
      char  Line [CGIPLUS_LINE_SIZE];
      FILE  *CgiPlusIn;
                                                 
      VarPtrSize = 0;
      if ((CgiPlusIn = fopen (getenv("CGIPLUSIN"), "r")) == NULL)
         return (NONE);
      while (fgets (Line, sizeof(Line), CgiPlusIn) != NULL)
      {
         /* first empty line signals the end of CGIplus variables */
         if (Line[0] == '\n') break;
         /* ignore any non-alphabetic line */
         if (!isalpha(Line[0])) continue;
         LineLength = strlen(Line);
         if ((VarPtr = realloc (VarPtr, VarPtrSize+LineLength+2)) == NULL)
            return (NONE);
         memcpy (VarPtr+VarPtrSize, Line, LineLength+1);
         VarPtrSize += LineLength;
         VarPtr[VarPtrSize-1] = '\0';
      }
      fclose (CgiPlusIn);
      if (VarPtr == NULL) return (NONE);
   }

   /* if just waiting for next request return now it's arrived */
   if (!VarName[0]) return (NONE);

   if (VarName[0] == '*')
   {
      /* return each stored line one at a time */
      if (NextPtr == NULL) NextPtr = VarPtr;
      if (!*(sptr = cptr = NextPtr))
      {
         NextPtr = NULL;
         return (NONE);
      }
      while (*cptr) cptr++;
      NextPtr = cptr + 1;
      return (sptr);
   }

   /* search CGIplus variable strings for required variable */
   sptr = VarPtr;
   while (*sptr)
   {
      /* simple comparison between supplied and in-string variable names */
      cptr = VarName;
      while (*cptr && *sptr && *sptr != '=')
      {
         if (toupper(*cptr) != toupper(*sptr)) break;
         cptr++;
         sptr++;
      }
      /* if found then return a pointer to the value */
      if (!*cptr && *sptr == '=') return (sptr+1);
      /* skip to next variable string (if any) */
      while (*sptr) sptr++;
      sptr++;
   }
   /* not found */
   return (NONE);
}

/*****************************************************************************/

