/*****************************************************************************/
/*
                                 Error.c

Error reporting functions.

All errors are reported using HTML formatting.  The assumptions being: 

  o  That most errors will occur before a document begins being transfered
     (e.g. file not found), and so its content type has not been set, and 
     the error can sent with a "text/html" type.

  o  HTML documents are probably more extensive than plain-text documents.
     Hence if an error occurs during transfer (e.g. record too big for user's
     buffer) it can be reported within the document with correct formatting 
     (document context notwithstanding).

  o  If an error occurs while transfering a plain-text document the HTML
     message is still readable, albeit a little cluttered.  Additional,
     obvious text is included to try and highlight the message.

  o  In non-"text/..." transfers the message text is not included anyway.

After a request thread is created and an error is generated the storage
'rqptr->ErrorMessagePtr' becomes non-NULL.  This mechanism can be used 
to detect whether an error has occured.  Most common errors create an error 
message in heap memory and point at this using 'rqptr->ErrorMessagePtr'. 
This message can be output at some time convenient to the task underway.  
After an error message is sent to the client the 'rqptr->ErrorMessagePtr' 
is returned to a NULL value to indicate no outstanding error (but usually the 
thread is disposed of after an error occurs).  Notification of other, more 
critical errors are sent directly to the client at the time they occur (for 
instance, failed heap memory allocation).

If ErrorGeneral() is called without first setting rqptr->ResponseStatusCode
it will default to 500, or "internal server error".  Therefore string
overflows and the like can be called without setting this code.  Query string
errors should set to 403 or 404 indication request cannot be processed.

If ErrorVmsStatus() is called without first setting rqptr->ResponseStatusCode
the function will attempt to determine the most appropriate HTTP status code
from the VMS status value.  It will then default to 500, or "internal server
error", indicating some unsual (e.g. non file system) condition.

The server configuration can supply additional information included whenever
an error report is generated.  This includes a URL for contacting the system
administrator (note: URL, not text, etc.) when an implementation,
configuration, process quota, etc., error is reported.   It also can provide
text to be included against any other error report.  This text can contain
HTML markup, carriage control, etc., and is designed for a brief message
containing a link to an explanation page.


VERSION HISTORY
---------------
01-FEB-97  MGD  HTTPd version 4
06-JUN-96  MGD  added authentication failure
01-DEC-95  MGD  HTTPd version 3
25-MAR-95  MGD  minor functional and cosmetic changes
20-DEC-94  MGD  multi-threaded daemon
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <string.h>

/* VMS related header files */
#include <descrip.h>
#include <iodef.h>
#include <ssdef.h>
#include <stsdef.h>

/* application related header files */
#include "wasd.h"
#include "error.h"
#include "httpd.h"
#include "net.h"
#include "support.h"

/******************/
/* global storage */
/******************/

/*
   IMPORTANT NOTE ... those with a fixed size [n] can have additional text
   appended to them.  Change storage capacility of these strings carefully!
*/
char  ErrorAuthorizationRequired [] =
      "Facility cannot be used without being authorization access controlled.",
      ErrorFacilityDisabled [] = "Facility disabled.",
      ErrorMemoryAllocation [40+64+128] = "Memory allocation failed.", 
      ErrorRedirectionCount [48+64+128] = "Redirection loop detected.",
      ErrorRequestBodyRead [] = "reading request content",
      ErrorRequestFormat [32+64] = "Cannot understand HTTP request.",
      ErrorRequestHttpMethod [] = "HTTP method not implemented.",
      ErrorRequestMaxAccept [] = "Maximum header <I>Accept:</I> exceeded.",
      ErrorSanityCheck [] = "Sanity check failure!",
      ErrorStringSize [24+64+128] = "String too small.",                      
      ErrorTimeFmt [24+64] = "Time format problem.",
      ErrorUrlEncoding [24+64] = "URL-encoding problem.";

char  *ErrorRecommendNotifyPtr;

/********************/
/* external storage */
/********************/

extern boolean  Debug;
extern char  HtmlSgmlDoctype[];
extern char  HttpProtocol[];
extern char  SoftwareID[];
extern char  Utility[];
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;

/*****************************************************************************/
/*
Called from HTTPd.c during startup.
*/

ErrorInitialize ()

{
   static char  RecommendBrowser [] =
                   "\n<P><I>(indicates browser problem)</I>",
                RecommendCorrect [] =
                   "\n<P><I>(document requires correction)</I>",
                RecommendNotify [] =
                   "\n<P><I>(notify system administrator)</I>";   

   static $DESCRIPTOR (NotifyUrlFaoDsc,
"\n<P><I>(notify <A HREF=\"!AZ\">system administrator</A>)</I>\0");

   static char  NotifyUrl [64+128];

   int  status;
   char  *ErrorNotifyPtr;
   $DESCRIPTOR (NotifyUrlDsc, NotifyUrl);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorInitialize()\n");

   if (Config.ErrorRecommend)
   {
      if (!Config.ErrorSysAdminPtr[0])
         ErrorRecommendNotifyPtr = RecommendNotify;
      else
      {
         status = sys$fao (&NotifyUrlFaoDsc, 0, &NotifyUrlDsc,
                           Config.ErrorSysAdminPtr);
         if (VMSnok (status) || status == SS$_BUFFEROVF)
            ErrorRecommendNotifyPtr = RecommendNotify;
         else
            ErrorRecommendNotifyPtr = NotifyUrl;
      }

      strcat (ErrorRequestFormat, RecommendBrowser);
      strcat (ErrorUrlEncoding, RecommendBrowser);

      strcat (ErrorTimeFmt, RecommendCorrect);

      strcat (ErrorMemoryAllocation, ErrorRecommendNotifyPtr);
      strcat (ErrorRedirectionCount, ErrorRecommendNotifyPtr);
      strcat (ErrorStringSize, ErrorRecommendNotifyPtr);
   }
}

/*****************************************************************************/
/*
Exit the application after output of the module and line location reporting 
the error, a brief explanation, and then exiting generating a DCL-reported 
message.
*/

ErrorExitVmsStatus
(
int VmsStatus,
char *Explanation,
char *SourceFileName,
int SourceLineNumber
)
{
   register char  *cptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorExitVmsStatus()\n");

   /* 
      The source file format provided by the "__FILE__" macro will
      be "device:[directory]name.type;ver".  Reduce that to "name".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   fprintf (stdout,
"%%%s-E-SOFTWAREID, %s\n\
-%s-E-WHERE, module: %s line: %d\n\
-%s-E-WHAT, %s\n",
            Utility, SoftwareID,
            Utility, cptr, SourceLineNumber,
            Utility, Explanation);

   if (!VmsStatus) VmsStatus = STS$K_ERROR | STS$M_INHIB_MSG;
   exit (VmsStatus);
}

/****************************************************************************/
/*
It is a fatal error to call this function without an error message.  If the
HTTP method is HEAD then generate an appropriate response header and send it.
If not HEAD and a text document, generate a reponse header and include the
error message as the body of the response.  Generates "Content-Length:" and
"Keep-Alive:" fields as required.   Allows error reports of any length.
*/

ErrorSendToClient
(
struct RequestStruct *rqptr,
void *AstFunctionPtr
)
{
   /* contains HTML, also enough to make it obvious in plain text */
   static char  ErrorMessageDivider [] =
"<P><FONT COLOR=\"#ff0000\"><B><HR>\n\
\n\
<!-- -- --   E R R O R   R E P O R T   -- -- -->\n\
\n";

   static $DESCRIPTOR (HttpErrorHeaderFaoDsc,
"!AZ !UL Error.\r\n\
Server: !AZ\r\n\
Content-Type: text/html\r\n\
Content-Length: !UL\r\n\
!AZ\
\r\n");

   static $DESCRIPTOR (HttpErrorAuthorizationFaoDsc,
"!AZ 401 Error.\r\n\
!AZ\
!AZ\
Server: !AZ\r\n\
ContentType: text/html\r\n\
Content-Length: !UL\r\n\
!AZ\
\r\n");

   static $DESCRIPTOR (BufferDsc, "");

   register unsigned long  *vecptr;

   int  status;
   unsigned short  Length;
   unsigned long  FaoVector [8];
   char  *KeepAlivePtr;
   char  Buffer [2048];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorSendToClient()\n");

   if (rqptr->ErrorMessagePtr == NULL || !rqptr->ErrorMessageLength)
      ErrorExitVmsStatus (0, ErrorSanityCheck, __FILE__, __LINE__);

   if (rqptr->BytesTx)
   {
      /*******************/
      /* provide divider */
      /*******************/

      /* buffer the original AST completion address */
      rqptr->ErrorMessageAstAddress = AstFunctionPtr;
      status = QioNetWrite (rqptr, &ErrorMessageDividerAst,
                            ErrorMessageDivider,
                            sizeof(ErrorMessageDivider)-1);
      return (status);
   }
   else
   {
      /***********************/
      /* provide HTTP header */
      /***********************/

      BufferDsc.dsc$a_pointer = Buffer;
      BufferDsc.dsc$w_length = sizeof(Buffer)-1;

      /*
         Mosaic 2.4-7 seems to spit on authentication if a persistent
         connection times-out (09-JUL-96)
      */
      if (rqptr->HttpUserAgentPtr != NULL &&
          strstr (rqptr->HttpUserAgentPtr, "Mosaic") != NULL)
         KeepAlivePtr = "";
      else
      if (Config.KeepAliveTimeoutSeconds &&
          rqptr->KeepAliveRequest &&
          (rqptr->HttpMethod == HTTP_METHOD_GET ||
           rqptr->HttpMethod == HTTP_METHOD_HEAD))
      {
         rqptr->KeepAliveCount++;
         rqptr->KeepAliveResponse = true;
         KeepAlivePtr = KeepAliveHttpHeader;
      }
      else
         KeepAlivePtr = "";

      if (rqptr->ResponseStatusCode == 401)
      {
         /*************************/
         /* authorization failure */
         /*************************/

         if (!rqptr->AuthChallengeScheme)
         {
            if (Config.AuthDigestEnabled)
               rqptr->AuthChallengeScheme |= AUTH_SCHEME_DIGEST;
            if (Config.AuthBasicEnabled)
               rqptr->AuthChallengeScheme |= AUTH_SCHEME_BASIC;
         }

         if (!rqptr->AuthChallengeScheme)
            ErrorExitVmsStatus (0, ErrorSanityCheck, __FILE__, __LINE__);

         if ((rqptr->AuthChallengeScheme & AUTH_SCHEME_DIGEST) &&
             rqptr->AuthDigestChallengePtr == NULL)
            DigestChallenge (rqptr, "");

         if ((rqptr->AuthChallengeScheme & AUTH_SCHEME_BASIC) &&
             rqptr->AuthBasicChallengePtr == NULL)
            BasicChallenge (rqptr);

         if (rqptr->AuthDigestChallengePtr == NULL)
            rqptr->AuthDigestChallengePtr = "";
         if (rqptr->AuthBasicChallengePtr == NULL)
            rqptr->AuthBasicChallengePtr = "";

         vecptr = FaoVector;
         *vecptr++ = HttpProtocol;
         *vecptr++ = rqptr->AuthDigestChallengePtr;
         *vecptr++ = rqptr->AuthBasicChallengePtr;
         *vecptr++ = SoftwareID;
         *vecptr++ = rqptr->ErrorMessageLength;
         *vecptr++ = KeepAlivePtr;

         status = sys$faol (&HttpErrorAuthorizationFaoDsc, &Length, &BufferDsc,
                            &FaoVector);
         if (VMSnok (status) || status == SS$_BUFFEROVF)
         {
            ErrorInternal (rqptr, status, "sys$faol()", __FILE__, __LINE__);
            return (status);
         }
      }
      else
      {
         /*****************/
         /* general error */
         /*****************/

         vecptr = FaoVector;
         *vecptr++ = HttpProtocol;
         *vecptr++ = rqptr->ResponseStatusCode;
         *vecptr++ = SoftwareID;
         *vecptr++ = rqptr->ErrorMessageLength;
         *vecptr++ = KeepAlivePtr;

         status = sys$faol (&HttpErrorHeaderFaoDsc, &Length, &BufferDsc,
                            &FaoVector);
         if (VMSnok (status) || status == SS$_BUFFEROVF)
         {
            ErrorInternal (rqptr, status, "sys$faol()", __FILE__, __LINE__);
            return (status);
         }
      }

      Buffer[BufferDsc.dsc$w_length = Length] = '\0';
      if (Debug) fprintf (stdout, "%d |%s|\n", Length, Buffer);

      if ((rqptr->ResponseHeaderPtr = HeapAlloc (rqptr, Length+1)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         return (SS$_INSFMEM);
      }
      memcpy (rqptr->ResponseHeaderPtr, Buffer, Length+1);
   }

   /**********************/
   /* send error message */
   /**********************/

   QioNetWrite (rqptr, AstFunctionPtr,
                rqptr->ErrorMessagePtr, rqptr->ErrorMessageLength);

   /* indicate the error message has been sent */
   rqptr->ErrorMessagePtr = NULL;
   rqptr->ErrorMessageLength = 0;

   return (SS$_NORMAL);
}

/*****************************************************************************/
/*
*/

ErrorMessageDividerAst (struct RequestStruct *rqptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorMessageDividerAst()\n");

   QioNetWrite (rqptr, rqptr->ErrorMessageAstAddress,
                rqptr->ErrorMessagePtr, rqptr->ErrorMessageLength);

   /* indicate the error message has been sent */
   rqptr->ErrorMessagePtr = NULL;
   rqptr->ErrorMessageLength = 0;
}

/*****************************************************************************/
/*
Generate an error message about a VMS status problem for subsequent reporting
to the client.
*/

ErrorVmsStatus
(
struct RequestStruct *rqptr,
int StatusValue,
char *SourceFileName,
int SourceLineNumber
)
{
   static char  DocumentNotFound [] = "Document not found",
                FileNotFound [] = "File not found",
                DocumentProtVio [] = "Document protection violation",
                FileProtVio [] = "File protection violation";

   static char  ErrorFao [] =
"!AZ\
<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"!AZ\">\n\
<META NAME=\"module\" CONTENT=\"!AZ\">\n\
<META NAME=\"line\" CONTENT=\"!UL\">\n\
<TITLE>Error !UL</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>ERROR!!</H1>\n\
<P>Reported by server.\n\
<P>!AZ ... <TT>!AZ</TT>\n\
<!!-- sts: %X!XL \"!AZ\" -->\n\
!AZ\
!AZ\
</BODY>\n\
</HTML>\n";
   static $DESCRIPTOR (ErrorFaoDsc, ErrorFao);

   register char  *cptr, *sptr;

   int  status;
   unsigned short  Length;
   char  *ContentTypePtr,
         *MessagePtr,
         *RecommendPtr;
   char  Message [256],
         Buffer [2048];
   $DESCRIPTOR (MessageDsc, Message);
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorVmsStatus()\n");

   /* don't overwrite any existing error message */
   if (rqptr->ErrorMessagePtr != NULL) return;

   if (rqptr->ContentType.ContentTypePtr == NULL)
      ContentTypePtr = "";
   else
      ContentTypePtr = rqptr->ContentType.ContentTypePtr;

   if (StatusValue == RMS$_FNF)
   {
      if (!rqptr->ResponseStatusCode)
         rqptr->ResponseStatusCode = 404;
      if (strsame (ContentTypePtr, "text/", 5))
         MessagePtr = DocumentNotFound;
      else
         MessagePtr = FileNotFound;
   }
   else
   if (StatusValue == RMS$_PRV)
   {
      if (!rqptr->ResponseStatusCode)
         rqptr->ResponseStatusCode = 403;
      if (strsame (ContentTypePtr, "text/", 5))
         MessagePtr = DocumentProtVio;
      else
         MessagePtr = FileProtVio;
   }
   else
   if (VMSok (status = sys$getmsg (StatusValue, &Length, &MessageDsc, 1, 0)))
   {
      if (!rqptr->ResponseStatusCode)
      {
         if (StatusValue == RMS$_DNF ||
             StatusValue == RMS$_SYN ||
             StatusValue == RMS$_DEV ||
             StatusValue == RMS$_FNM ||
             StatusValue == RMS$_TYP ||
             StatusValue == RMS$_DIR ||
             StatusValue == SS$_NOSUCHFILE)
            rqptr->ResponseStatusCode = 400;
         else
         /* could be generated by Put.c */
         if (StatusValue == SS$_DIRNOTEMPTY ||
             StatusValue == SS$_EXDISKQUOTA)
            rqptr->ResponseStatusCode = 400;
      }
      Message[Length] = '\0';
      *(MessagePtr = Message) = *Message;
      cptr = sptr = Message;
      *cptr = toupper(*cptr);
      /* improve the look by removing any embedded sys$fao() formatting */
      while (*cptr)
      {
         if (*cptr == '!')
         {
            cptr++;
            *sptr++ = '?';
            /* step over any field width digits */
            while (isdigit(*cptr)) cptr++;
            /* usually two formatting characters */
            if (isalpha(*cptr)) cptr++;
            if (isalpha(*cptr)) cptr++;
         }
         else
            *sptr++ = *cptr++;
      }
      *sptr = '\0';
      /** if (Debug) fprintf (stdout, "Message |%s|\n", Message); **/
   }
   else
   {
      ErrorInternal (rqptr, status, "sys$fao()", __FILE__, __LINE__);
      return;
   }

   if (!rqptr->ResponseStatusCode) rqptr->ResponseStatusCode = 500;

   if (rqptr->ErrorTextPtr == NULL)
      if (rqptr->PathInfoPtr == NULL)
         rqptr->ErrorTextPtr = "(no information available)";
      else
         rqptr->ErrorTextPtr = rqptr->PathInfoPtr;

   if (rqptr->ErrorHiddenTextPtr == NULL || !Config.IncludeCommentedInfo)
      rqptr->ErrorHiddenTextPtr = "(no information available)";

   if (!Config.ErrorRecommend || rqptr->ResponseStatusCode == 401)
      RecommendPtr = "";
   else
   if (StatusValue == RMS$_FNF ||
       StatusValue == RMS$_DNF ||
       StatusValue == SS$_NOSUCHFILE)
      RecommendPtr = "<P><I>(document, or bookmark, requires revision)</I>\n";
   else
   if (StatusValue == RMS$_PRV)
      RecommendPtr = "<P><I>(protection requires revision)</I>\n";
   else
   if (StatusValue == SS$_NOPRIV)
      RecommendPtr = "<P><I>(no authorization to access the object)</I>\n";
   else
   if (StatusValue == RMS$_SYN ||
       StatusValue == RMS$_DEV ||
       StatusValue == RMS$_DIR ||
       StatusValue == RMS$_FNM ||
       StatusValue == RMS$_TYP)
      RecommendPtr =
         "<P><I>(document, or specification, requires correction)</I>\n";
   else
   if (StatusValue == RMS$_FLK)
      RecommendPtr = "<P><I>(try again shortly)</I>\n";
   else
   if (StatusValue == SS$_DIRNOTEMPTY ||
       StatusValue == SS$_EXDISKQUOTA)
      RecommendPtr = "<P><I>(correct situation, try again)</I>\n";
   else
      RecommendPtr = ErrorRecommendNotifyPtr;

   /* 
      The source file format provided by the "__FILE__" macro will
      be "device:[directory]name.type;ver".  Reduce that to "name".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      sptr = cptr;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   status = sys$fao (&ErrorFaoDsc, &Length, &BufferDsc,
                     HtmlSgmlDoctype, SoftwareID, cptr, SourceLineNumber,
                     rqptr->ResponseStatusCode,
                     MessagePtr, rqptr->ErrorTextPtr,
                     StatusValue, rqptr->ErrorHiddenTextPtr,
                     RecommendPtr, Config.ErrorInfoPtr);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      ErrorInternal (rqptr, status, "sys$fao()", __FILE__, __LINE__);
      return;
   }

   if ((rqptr->ErrorMessagePtr = HeapAlloc (rqptr, Length+1)) == NULL)
   {
      ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
      return;
   }

   memcpy (rqptr->ErrorMessagePtr, Buffer, Length);
   rqptr->ErrorMessagePtr[rqptr->ErrorMessageLength = Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", rqptr->ErrorMessagePtr);
}

/*****************************************************************************/
/*
Generate an error message about a general (non-VMS status) problem for
subsequent reporting to the client.
*/
 
ErrorGeneral
(
struct RequestStruct *rqptr,
char *Explanation,
char *SourceFileName,
int SourceLineNumber
)
{
   static $DESCRIPTOR (ErrorFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"!AZ\">\n\
<META NAME=\"module\" CONTENT=\"!AZ\">\n\
<META NAME=\"line\" CONTENT=\"!UL\">\n\
<TITLE>Error !UL</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>ERROR!!</H1>\n\
<P>Reported by server.\n\
<P>!AZ\n\
!AZ\
</BODY>\n\
</HTML>\n");

   register unsigned char  *cptr;

   int  status;
   unsigned short  Length;
   char  Buffer [2048];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorGeneral()\n");

   /* don't overwrite any existing error message */
   if (rqptr->ErrorMessagePtr != NULL) return;

   /*
      The source file format provided by the "__FILE__" macro will
      be "device:[directory]name.type;ver".  Reduce that to "name".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (!rqptr->ResponseStatusCode) rqptr->ResponseStatusCode = 500;

   status = sys$fao (&ErrorFaoDsc, &Length, &BufferDsc,
                     HtmlSgmlDoctype, SoftwareID, cptr, SourceLineNumber,
                     rqptr->ResponseStatusCode, Explanation,
                     Config.ErrorInfoPtr);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      ErrorInternal (rqptr, status, "sys$fao()", __FILE__, __LINE__);
      return;
   }

   if ((rqptr->ErrorMessagePtr = HeapAlloc (rqptr, Length+1)) == NULL)
   {
      ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
      return;
   }

   memcpy (rqptr->ErrorMessagePtr, Buffer, Length);
   rqptr->ErrorMessagePtr[rqptr->ErrorMessageLength = Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", rqptr->ErrorMessagePtr);
}

/*****************************************************************************/
/*
*/
 
ErrorInternal
(
struct RequestStruct *rqptr,
int Status,
char *Explanation,
char *SourceFileName,
int SourceLineNumber
)
{
   static $DESCRIPTOR (BufferFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
<META NAME=\"generator\" CONTENT=\"!AZ\">\n\
<META NAME=\"module\" CONTENT=\"!AZ\">\n\
<META NAME=\"line\" CONTENT=\"!UL\">\n\
<HEAD>\n\
<TITLE>Error 500</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>ERROR!!</H1>\n\
<P>Reported by server.\n\
<P><FONT COLOR=\"#ff0000\">SERVER INTERNAL ERROR!!</FONT> \
... !AZ <!!-- %X!XL -->\n\
!AZ\
!AZ\
</BODY>\n\
</HTML>\n");

   register char  *cptr;
   unsigned short  Length;
   char  *RecommendPtr;
   char  Buffer [1024];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorInternal()\n");

   Accounting.InternalErrorCount++;

   rqptr->ResponseStatusCode = 500;

   /* 
      The source file format provided by the "__FILE__" macro will
      be "device:[directory]name.type;ver".  Reduce that to "name".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (!Config.ErrorRecommend)
      RecommendPtr = "";
   else
      RecommendPtr = ErrorRecommendNotifyPtr;

   sys$fao (&BufferFaoDsc, &Length, &BufferDsc,
            HtmlSgmlDoctype, SoftwareID, cptr, SourceLineNumber,
            Explanation, Status, RecommendPtr, Config.ErrorInfoPtr);
   Buffer[rqptr->ErrorMessageLength = Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", Buffer);
   if ((rqptr->ErrorMessagePtr = HeapAlloc (rqptr, Length+1)) == NULL)
   {
      ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
      return;
   }
   memcpy (rqptr->ErrorMessagePtr, Buffer, Length+1);
}

/*****************************************************************************/
/*
Report heap memory allocation problem IMMEDIATELY to the client (PS: can't use
dynamic memory to report, so do it in two parts, one that waits so that
automatic storage can be used, the second a static message).
*/                     
 
ErrorHeapAlloc
(
struct RequestStruct *rqptr,
char *SourceFileName,
int SourceLineNumber
)
{
   static char  FaoFailed [] =
"HTTP/1.0 500 Error.\r\n\
\r\n\
<HTML>\n\
<HEAD>\n\
<TITLE>Error 500</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>ERROR!</H1>\n\
<P>Reported by server.\n\
<P>sys$fao() failed in HeapAlloc().\n\
</BODY>\n\
</HTML>\n";


   static $DESCRIPTOR (ErrorFaoDsc,
"!AZ 500 Error.\r\n\
Server: !AZ\r\n\
\r\n\
<!!-- SoftwareID: !AZ Module: !AZ Line: !UL -->\n\
<HTML>\n\
<HEAD>\n\
<TITLE>Error 500</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>ERROR!!</H1>\n\
<P>Reported by server.\n\
<P>Heap allocation failed.\n\
!AZ\
!AZ\
</BODY>\n\
</HTML>\n");

   register char  *cptr;

   int  status;
   unsigned short  Length;
   char  *RecommendPtr;
   char  Buffer [1024];
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "ErrorHeapAlloc()\n");

   rqptr->ResponseStatusCode = 500;

   /* 
      The source file format provided by the "__FILE__" macro will
      be "device:[directory]name.type;ver".  Reduce that to "name".
   */
   for (cptr = SourceFileName; *cptr && *cptr != ';'; cptr++);
   if (*cptr)
   {
      while (*cptr != '.') cptr--;
      *cptr-- = '\0';
   }
   while (*cptr != ']') cptr--;
   cptr++;

   if (!Config.ErrorRecommend)
      RecommendPtr = "";
   else
      RecommendPtr = ErrorRecommendNotifyPtr;

   status = sys$fao (&ErrorFaoDsc, &Length, &BufferDsc,
                     SoftwareID, cptr, SourceLineNumber, RecommendPtr,
                     Config.ErrorInfoPtr);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      NetWrite (rqptr->ClientChannel, FaoFailed, sizeof(FaoFailed)-1);
      return;
   }
   Buffer[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", Buffer);

   NetWrite (rqptr->ClientChannel, Buffer, Length);
}

/*****************************************************************************/
/*
*/
 
char* SysGetMsg (int StatusValue)
 
{
   static char  Message [256];
   short int  Length;
   $DESCRIPTOR (MessageDsc, Message);
 
   sys$getmsg (StatusValue, &Length, &MessageDsc, 0, 0);
   Message[Length] = '\0';
   if (Debug) fprintf (stdout, "SysGetMsg() |%s|\n", Message);
   return (Message);
}
 
/****************************************************************************/

