/*****************************************************************************/
/*
                                 MapUrl.c

Functions supporting mapping of URLs to VMS file specifications and VMS 
specifications to URLs, uses the HTTPd rule mapping file.  The number of rules 
it can store is not fixed, but is limited by available memory and the 
practical considerations of processing large, linear rule databases. 

This module is designed to be completely self-contained, only the global 
storage listed below is required for linking.  This is to allow other 
programs, particularly CGI scripts, to use the URL mapping and reverse-mapping 
functionality if they require it.  That is the primary reason for the separate 
rule-mapping database file with the WASD VMS HTTPd.


VERSION HISTORY
---------------
05-JUN-97  MGD  version 4.2, CGIplus "exec+" and "script+" rules
01-FEB-97  MGD  HTTPd version 4;
                major changes (and a general rework) for form-based config
01-OCT-96  MGD  not proud of how I have shoe-horned the reporting features in,
                a good example of slop-down or object-disoriented programming,
                :^( will rework the whole thing one day, sigh!
                bugfix; requiring use of 'TheFirstRuleWillBeTheNextRule'
06-APR-96  MGD  modified conversion of non-VMS characters in MapUrl_UrlToVms()
15-FEB-95  MGD  bugfix; redirect rule
01-DEC-95  MGD  HTTPd version 3
24-MAR-95  MGD  bugfix; end-of-string sometimes not detected when rule matching
20-DEC-94  MGD  revised for multi-threaded HTTP daemon
20-JUN-94  MGD  single-threaded daemon
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

/* VMS related header files */
#include <prvdef.h>
#include <rms.h>
#include <ssdef.h>
#include <stsdef.h>

/* application header files */
#include "wasd.h"
#include "mapurl.h"

/********************************/
/* a little preprocessor config */
/********************************/

#undef IP_MAPURL
#ifdef IP_UCX
#   define IP_MAPURL 1
#endif /* IP_UCX */
#ifdef IP_MULTINET
#   define IP_MAPURL 1
#endif /* IP_MULTINET */

/**********/
/* macros */
/**********/

#define boolean int
#define true 1
#define false 0
 
#define VMSok(x) ((x) & STS$M_SUCCESS)
#define VMSnok(x) !(((x) & STS$M_SUCCESS))

/*******************/
/* data structures */
/*******************/

#ifdef __ALPHA
#   pragma member_alignment __save
#   pragma member_alignment
#endif

/* rule database is a singly-linked list of these structures */
struct MappingRuleStruct
{
   struct MappingRuleStruct  *NextPtr;
   boolean  IsCgiPlusScript;
   int  MapFileLineNumber;
   char  RuleType;
   char  *TemplatePtr;
   char  *ResultPtr;
   /* template and result null-terminated strings stored straight after */
};

struct MappingRuleListStruct
{
   struct MappingRuleStruct  *HeadPtr;
   struct MappingRuleStruct  *TailPtr;
   int  LineNumber,
        ProblemCount,
        ProblemReportLength;
   unsigned long  LoadBinTime [2],
                  RevBinTime [2];
   char  *LinePtr,
         *ProblemReportPtr;
   char  LoadFileName [256],
         RuleName [32];
   /* only used when part of an HTTP request */
   struct RequestStruct  *RequestPtr;
};

#ifdef __ALPHA
#   pragma member_alignment __restore
#endif

/******************/
/* global storage */
/******************/

struct MappingRuleListStruct  MappingRulesList;

/*
   Point at the above storage as default for a rules list.
   This global pointer may be changed temporarily for using other rules.
*/
struct MappingRuleListStruct  *GlobalMappingRulesListPtr = &MappingRulesList;

/********************/
/* external storage */
/********************/

extern boolean  Debug;
extern char  SoftwareID[];
extern char  Utility[];

#ifdef IP_MAPURL

#include "support.h"

extern int  ServerPort;
extern char  ErrorStringSize[];
extern char  ErrorUrlEncoding[];
extern char  HtmlSgmlDoctype[];
extern char  ServerHostName[];

#endif /* IP_MAPURL */

/*****************************************************************************/
/*
This overly-long function (sorry) is not AST reentrant, but when executed 
within the context of the HTTPd, is executed either without the possibility of 
AST interruption, or during AST processing and so provides atomic 
functionality.  Pointers to any strings returned must be used immediately 
(before leaving AST context), or copied into storage maintained by the calling 
routine. 

Maps from URL format to VMS file specification, and REVERSE-MAPS from VMS file 
specification to URL format.  Maps scripts.

Always returns a pointer to char.  An error message is detected by being  
returned as a pointer to a string beginning with a null-character!  The 
message begins from character one. 

Call with 'PathPtr' pointing at a string containing the URL path and 'VmsPtr' 
pointing to the storage to contain the mapped VMS equivalent.

Call with 'VmsPtr' pointing to a VMS file specification (dev:[dir]file.ext) 
and 'PathPtr' pointing to an empty string, used as storage to contain the 
REVERSE-MAPPED file specification.

Call with 'PathPtr' pointing at a string containing the URL path, 'VmsPtr' 
pointing to storage to contain the mapped VMS equivalent, 'ScriptPtr' pointing 
to storage for the URL format script path and 'ScriptVmsPtr' pointing to 
storage to contain the VMS specification of the script procedure/image.  If 
there are two consecutive forward slashes following a script specification the 
path derived from the characters following the script are not mapped to a VMS 
specification, it is returned as is.  Allows "raw" paths to be provided to 
scripts.

Current Rules:

EXEC       template   result
FAIL       template
MAP        template   result
PASS       template   [result]
REDIRECT   template   result
SCRIPT     template   result
*/ 

char* MapUrl
(
char *PathPtr,
char *VmsPtr,
char *ScriptPtr,
char *ScriptVmsPtr
)
{
   static char  AccessDeniedDefault [] = 
                "\0Access denied (forbidden by default).",
                AccessDeniedInternal [] =
                "\0Access denied (INTERNAL processing problem!)",
                AccessDeniedNoRules [] = 
                "\0Access denied (NO RULES LOADED!)",
                AccessDeniedRule [] = 
                "\0Access denied (forbidden by rule).";

   static char  DerivedPathBuffer [512],
                PathBuffer [512],
                VmsBuffer [512];

   static struct MappingRuleStruct  TheFirstRuleWillBeTheNextRule;

   register char  PathChar,
                  RuleChar,
                  LastNonWildcardPathChar,
                  LastNonWildcardRuleChar;
   register char  *cptr, *pptr, *sptr, *zptr;
   register struct MappingRuleStruct  *mrptr;

   boolean  LoadingRules,
            MapPathToVms;
   int  status,
        TotalLength;
   unsigned short  Length;
   char  Scratch [512],
         WildBuffer [512];

   /*********/
   /* begin */
   /*********/

   if (Debug)
   {
      fprintf (stdout, "MapUrl()");
      if (PathPtr != NULL) fprintf (stdout, " PathPtr: |%s|", PathPtr);
      if (VmsPtr != NULL) fprintf (stdout, " VmsPtr: |%s|", VmsPtr);
      if (ScriptPtr != NULL) fprintf (stdout, " (script)");
      fputc ('\n', stdout);
   }

   /* if just loading the rules then free any previous set */
   if (LoadingRules = (PathPtr == NULL && VmsPtr == NULL && ScriptPtr == NULL))
      MapUrl_FreeRules (GlobalMappingRulesListPtr);

   if (GlobalMappingRulesListPtr->HeadPtr == NULL)
   {
      /*********************/
      /* read mapping file */
      /*********************/

      cptr = MapUrl_LoadRules (GlobalMappingRulesListPtr);
      /* return message if an error reported */
      if (!cptr[0] && cptr[1]) return (cptr);
      /* point at the location of the first rule as the next rule */
      TheFirstRuleWillBeTheNextRule.NextPtr =
         GlobalMappingRulesListPtr->HeadPtr;
   }

   /* if just (re)loading rules then return now */
   if (LoadingRules) return ("\0\0");

   /*****************************************/
   /* map from the URL or VMS to VMS or URL */
   /*****************************************/

   if (PathPtr == NULL) *(PathPtr = PathBuffer) = '\0';
   if (VmsPtr == NULL) *(VmsPtr = VmsBuffer) = '\0';
   /* making 'ScriptPtr' non-empty means "exec" and "script" are ignored */
   if (ScriptPtr == NULL) ScriptPtr = "?";

   if (PathPtr[0])
   {
      /* if the URL is not an empty string then force the conversion to VMS */
      MapPathToVms = true;
      *VmsPtr = '\0';
   }
   else
   {
      /* URL was an empty string, convert from VMS to URL */
      MapPathToVms = false;
      /* generate a URL-style version of the VMS specification */
      MapUrl_VmsToUrl (PathPtr, VmsPtr);
   }

   /***********************/
   /* loop thru the rules */
   /***********************/

   if ((mrptr = GlobalMappingRulesListPtr->HeadPtr) == NULL)
      return (AccessDeniedNoRules);

   /*
      This loop may be restarted by the "exec" or "script" rule after mapping
      the script component to further map the derived path.  This is done by
      setting up 'PathPtr' to the derived path and then repointing 'mrptr'
      to a scratch structure which has its 'NextPtr' set up to the first rule
      in the list and 'continue'ing the loop.
   */

   for (/* above! */; mrptr != NULL; mrptr = mrptr->NextPtr)
   {
      if (Debug)
         fprintf (stdout, "Rule |%c|%s|%s|\n",
                  mrptr->RuleType, mrptr->TemplatePtr, mrptr->ResultPtr);

      /* when mapping from VMS to URL then ignore all BUT "pass" rules */
      if (!MapPathToVms && mrptr->RuleType != 'P') continue;

      /* if 'ScriptPtr' is not to be used then ignore "exec" and "script" */
      if (ScriptPtr[0] && (mrptr->RuleType == 'E' || mrptr->RuleType == 'S'))
         continue;

      /*****************************************************************/
      /* compare the URL with the template/result in the mapping entry */
      /*****************************************************************/

      cptr = mrptr->TemplatePtr;
      /* if reverse-mapping VMS to URL use the result if available */
      if (!MapPathToVms && mrptr->ResultPtr[0]) cptr = mrptr->ResultPtr;

      /* 'WildBuffer' contains the null-separated wildcard-matched portions */
      zptr = (sptr = WildBuffer) + sizeof(WildBuffer)-1;
      LastNonWildcardPathChar = PathChar = *(pptr = PathPtr);
      LastNonWildcardRuleChar = RuleChar = *cptr;
      /** if (Debug) fprintf (stdout, "|%s|%s|\n", pptr, cptr); **/

      while (PathChar && RuleChar &&
             (tolower(PathChar) == tolower(RuleChar) || RuleChar == '*'))
      {
         if (RuleChar != '*')
         {
            LastNonWildcardPathChar = PathChar;
            LastNonWildcardRuleChar = RuleChar;
            PathChar = *++pptr;
            RuleChar = *++cptr;
            /** if (Debug) fprintf (stdout, "|%s|%s|\n", pptr, cptr); **/
            continue;
         }
         /** if (Debug) fprintf (stdout, "|%s|%s|\n", pptr, cptr); **/
         if ((mrptr->RuleType != 'S' && mrptr->RuleType != 'E') ||
             (mrptr->RuleType == 'S' || mrptr->RuleType == 'E') &&
             (tolower(LastNonWildcardPathChar) ==
              tolower(LastNonWildcardRuleChar) || !*pptr))
         {
            /* asterisk wildcard, multiple consecutive treated as single */
            while ((RuleChar = *cptr) == '*') cptr++;
            if (PathChar) *sptr++ = PathChar;
            if (sptr >= zptr) return (AccessDeniedInternal);
            while ((PathChar = *++pptr) &&
                   tolower(PathChar) != tolower(RuleChar) &&
                   sptr < zptr)
               *sptr++ = PathChar;
            if (sptr >= zptr) return (AccessDeniedInternal);
            *sptr++ = '\0';
            if (sptr >= zptr) return (AccessDeniedInternal);
            /** if (Debug) fprintf (stdout, "|%s|%s|\n", pptr, cptr); **/
         }
         else
            while ((RuleChar = *cptr) == '*') cptr++;
      }

      /* account for trailing asterisk on rule template */
      if (RuleChar == '*')
         while ((RuleChar = *++cptr) == '*');

      /* if not matched then straight to next rule */
      if (PathChar || RuleChar)
      {
#ifdef IP_MAPURL
         if (GlobalMappingRulesListPtr->RequestPtr)
            MapUrl_ReportRule (mrptr, PathPtr, false);
#endif /* IP_MAPURL */

         continue;
      }
#ifdef IP_MAPURL
      else
      {
         if (GlobalMappingRulesListPtr->RequestPtr)
            MapUrl_ReportRule (mrptr, PathPtr, true);
      }
#endif /* IP_MAPURL */

      /* final, terminating null for 'WildBuffer' */
      *sptr = '\0';

      /****************/
      /* matched them */
      /****************/

      if (MapPathToVms)
      {
         /***************************/
         /* mapping from URL to VMS */
         /***************************/

         switch (mrptr->RuleType)
         {
         case 'E' :

            /* exec directory mapping rule (different to "script") */

            /* get the virtual path to the script (CGI: "WWW_SCRIPT_NAME") */
            cptr = WildBuffer;
            sptr = mrptr->TemplatePtr;
            zptr = (pptr = ScriptPtr) + 255;
            while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
            if (pptr >= zptr) return (AccessDeniedInternal);
            /* get the wildcard buffer up to the first slash (script name) */
            while (*cptr && *cptr != '/' && pptr < zptr) *pptr++ = *cptr++;
            if (pptr >= zptr) return (AccessDeniedInternal);
            *pptr = '\0';

            /* now get the URL-format path to the physical script */
            cptr = WildBuffer;
            sptr = mrptr->ResultPtr;
            zptr = (pptr = Scratch) + 255;
            /* get fixed portion of rule as script directory location */
            while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
            if (pptr >= zptr) return (AccessDeniedInternal);
            /* get the wildcard buffer up to the first slash (script name) */
            while (*cptr && *cptr != '/' && pptr < zptr) *pptr++ = *cptr++;
            if (pptr >= zptr) return (AccessDeniedInternal);
            *pptr = '\0';

            /* convert the URL-format script to a VMS-format specification */
            MapUrl_UrlToVms (Scratch, ScriptVmsPtr);

            /* indicate CGIplus script via leading '+' instead of '/' */
            if (mrptr->IsCgiPlusScript) ScriptPtr[0] = '+';

            /* get all including and after the first slash as the new path */
            zptr = (pptr = DerivedPathBuffer) + sizeof(DerivedPathBuffer)-1;
            while (*cptr && pptr < zptr) *pptr++ = *cptr++;
            if (pptr >= zptr) return (AccessDeniedInternal);
            *pptr++ = '\0';
            TotalLength = pptr - DerivedPathBuffer;

            if (Debug)
               fprintf (stdout, "EXEC |%s|%s|%s|\n",
                        ScriptPtr, ScriptVmsPtr, DerivedPathBuffer);

            if (!DerivedPathBuffer[0]) return ("\0\0");

            /* two consecutive slashes means do not map derived path */
            if (*((unsigned short*)DerivedPathBuffer) == 0x2f2f)
            {
               /* remove the leading slash, then return */
               memcpy (DerivedPathBuffer, DerivedPathBuffer+1, TotalLength-1);
               return (DerivedPathBuffer);
            }

            /* restarting at first rule map the path derived from the script */
            *VmsPtr = '\0';
            PathPtr = DerivedPathBuffer;
            mrptr = &TheFirstRuleWillBeTheNextRule;
            continue;

         case 'F' :

            if (Debug) fprintf (stdout, "FAIL |%s|\n", PathPtr);
            return (AccessDeniedRule);

         case 'M' :

            /* map the result onto the original (even for "redirect") */

            cptr = WildBuffer;
            /* place this in a buffer so not to overwrite original path */
            zptr = (pptr = PathPtr = PathBuffer) + sizeof(PathBuffer)-1;
            sptr = mrptr->ResultPtr;
            /* scan through the result string */
            while (*sptr)
            {
               while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
               if (pptr >= zptr) return (AccessDeniedInternal);
               if (!*sptr) break;
               /* a wildcard asterisk, substitute from original path */
               while (*++sptr == '*');
               while (*cptr && pptr < zptr) *pptr++ = *cptr++;
               if (pptr >= zptr) return (AccessDeniedInternal);
               cptr++;
            }
            *pptr = '\0';

            if (mrptr->RuleType == 'M')
            {
               /* continue with the substituted URL mapping */
               if (Debug) fprintf (stdout, "MAP |%s|\n", PathPtr);
               continue;
            }

            if (Debug) fprintf (stdout, "REDIRECT |%s|\n", PathPtr);
            return (PathPtr);

         case 'P' :
         case 'R' :

            /* if there is no result to map just pass back the original path */
            if (!*(sptr = mrptr->ResultPtr)) 
            {
               /* convert the URL-style to a VMS-style specification */
               MapUrl_UrlToVms (PathPtr, VmsPtr);

               if (Debug) fprintf (stdout, "PASS |%s|%s|\n", PathPtr, VmsPtr);
               return (PathPtr);
            }

            cptr = WildBuffer;
            zptr = (pptr = PathBuffer) + sizeof(PathBuffer)-1;
            /* scan through the result string */
            while (*sptr)
            {
               while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
               if (pptr >= zptr) return (AccessDeniedInternal);
               if (!*sptr) break;
               /* a wildcard asterisk, substitute from original path */
               while (*++sptr == '*');
               while (*cptr && pptr < zptr) *pptr++ = *cptr++;
               if (pptr >= zptr) return (AccessDeniedInternal);
               cptr++;
            }
            *pptr = '\0';

            if (mrptr->RuleType == 'P')
            {
               /* convert the URL-style to a VMS-style specification */
               MapUrl_UrlToVms (PathBuffer, VmsPtr);
               if (Debug) fprintf (stdout, "PASS |%s|%s|\n", PathPtr, VmsPtr);
               return (PathPtr);
            }
            else
            {
               if (Debug) fprintf (stdout, "PASS |%s|%s|\n", PathPtr, VmsPtr);
               return (PathBuffer);
            }


         case 'S' :

            /* script file name mapping rule (different to "exec") */

            /* get the virtual path to the script (CGI: "WWW_SCRIPT_NAME") */
            sptr = mrptr->TemplatePtr;
            zptr = (pptr = ScriptPtr) + 255;
            while (*sptr && *sptr != '*' && pptr < zptr) *pptr++ = *sptr++;
            if (pptr >= zptr) return (AccessDeniedInternal);
            if (pptr > ScriptPtr)
               if (pptr[-1] == '/')
                  pptr--;
            *pptr = '\0';

            /* now get the URL-format path to the physical script */
            zptr = (pptr = Scratch) + 255;
            sptr = mrptr->ResultPtr;
            /* 0x2a2f is the numeric equivalent of "/*" */
            while (*sptr && *sptr != '*' &&
                   *(unsigned short*)sptr != 0x2a2f && pptr < zptr)
               *pptr++ = *sptr++;
            if (pptr >= zptr) return (AccessDeniedInternal);
            *pptr = '\0';

            /* convert the URL-format script to a VMS-format specification */
            MapUrl_UrlToVms (Scratch, ScriptVmsPtr);

            /* indicate CGIplus script via leading '+' instead of '/' */
            if (mrptr->IsCgiPlusScript) ScriptPtr[0] = '+';

            /* get wildcard matched second section of path as new path */
            zptr = (pptr = DerivedPathBuffer) + sizeof(DerivedPathBuffer)-1;
            if (*sptr == '/') *pptr++ = '/';
            cptr = WildBuffer;
            while (*cptr && pptr < zptr) *pptr++ = *cptr++;
            if (pptr >= zptr) return (AccessDeniedInternal);
            *pptr++ = '\0';
            TotalLength = pptr - DerivedPathBuffer;

            if (Debug)
               fprintf (stdout, "SCRIPT |%s|%s|%s|\n",
                        ScriptPtr, ScriptVmsPtr, DerivedPathBuffer);

            if (!DerivedPathBuffer[0]) return ("\0\0");

            /* two consecutive slashes means do not map derived path */
            if (*((unsigned short*)DerivedPathBuffer) == 0x2f2f)
            {
               /* remove the leading slash, then return */
               memcpy (DerivedPathBuffer, DerivedPathBuffer+1, TotalLength-1);
               return (DerivedPathBuffer);
            }

            /* restarting at first rule map the path derived from the script */
            *VmsPtr = '\0';
            PathPtr = DerivedPathBuffer;
            mrptr = &TheFirstRuleWillBeTheNextRule;
            continue;
         }
      }
      else
      {
         /***************************/
         /* mapping from VMS to URL */
         /***************************/

         /* REVERSE maps a VMS "result" to a "template" :^) */

         cptr = WildBuffer;
         pptr = PathPtr;
         sptr = mrptr->TemplatePtr;
         /* scan through the template string */
         while (*sptr)
         {
            while (*sptr && *sptr != '*') *pptr++ = *sptr++;
            if (!*sptr) break;
            /* a wildcard asterisk, substitute from result path */
            while (*++sptr == '*');
            if (*cptr)
            {
               while (*cptr) *pptr++ = *cptr++;
               cptr++;
            }
         }
         *pptr = '\0';

         if (Debug) fprintf (stdout, "PASS |%s|\n", PathPtr);
         return (PathPtr);
      }
   }

   /***********************/
   /* a mapping not found */
   /***********************/

   return (AccessDeniedDefault);
}

/*****************************************************************************/
/*
Open the mapping rule file and reading each line place mapping rules into the
singly-linked list pointed to by the supplied list pointer.  Close the file.
*/ 
 
char* MapUrl_LoadRules (struct MappingRuleListStruct *rlptr)

{
   static unsigned long  SysPrvMask [2] = { PRV$M_SYSPRV, 0 };

   static int  MappingRuleStructOverhead = sizeof(struct MappingRuleStruct);

   static char  ProblemConfused [] = "confused",
                ProblemResultRequired [] = "requires a \"result\"",
                ProblemTemplateRequired [] = "requires a \"template\"",
                ProblemWildcardMapping [] = "wildcard mapping problem";

   static char  ErrorString [256];

   register char  RuleChar;
   register char  *cptr, *sptr, *zptr;
   register struct MappingRuleStruct  *mrptr;

   boolean  IsCgiPlusScript,
            RequiresResult;
   int  status,
        SetPrvStatus,
        TemplateLength,
        TemplateWildcardCount,
        TotalLength,
        ResultWildcardCount;
   unsigned short  Length;
   char  *ResultPtr;
   char  ExpandedFileName [256],
         Line [256],
         Scratch [256];
   $DESCRIPTOR (ErrorStringDsc, ErrorString);
   struct FAB  MapFileFab;
   struct NAM  MapFileNam;
   struct RAB  MapFileRab;
   struct XABDAT  MapFileXabDat;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MapUrl_LoadRules() |%s|\n", MapFileName);

   MapUrl_FreeRules (rlptr);

   if (rlptr->ProblemReportPtr != NULL) free (rlptr->ProblemReportPtr);

   memset (rlptr, 0, sizeof(struct MappingRuleListStruct));

   sys$gettim (&rlptr->LoadBinTime);

   MapFileFab = cc$rms_fab;
   MapFileFab.fab$b_fac = FAB$M_GET;
   MapFileFab.fab$l_fna = MapFileName;
   MapFileFab.fab$b_fns = strlen(MapFileName);
   MapFileFab.fab$l_nam = &MapFileNam;
   MapFileFab.fab$b_shr = FAB$M_SHRGET;
   MapFileFab.fab$l_xab = &MapFileXabDat;

   MapFileNam = cc$rms_nam;
   MapFileNam.nam$l_esa = ExpandedFileName;
   MapFileNam.nam$b_ess = sizeof(ExpandedFileName)-1;
   MapFileNam.nam$l_rsa = rlptr->LoadFileName;
   MapFileNam.nam$b_rss = sizeof(rlptr->LoadFileName)-1;

   /* initialize the date extended attribute block */
   MapFileXabDat = cc$rms_xabdat;

   /* turn on SYSPRV to allow access to possibly protected file */
   if (VMSnok (SetPrvStatus = sys$setprv (1, &SysPrvMask, 0, 0)))
#ifdef IP_MAPURL
      ErrorExitVmsStatus (SetPrvStatus, "sys$setprv()", __FILE__, __LINE__);
#else
      exit (SetPrvStatus);
#endif

   status = sys$open (&MapFileFab, 0, 0);

   /* turn off SYSPRV */
   if (VMSnok (SetPrvStatus = sys$setprv (0, &SysPrvMask, 0, 0)))
#ifdef IP_MAPURL
      ErrorExitVmsStatus (SetPrvStatus, "sys$setprv()", __FILE__, __LINE__);
#else
      exit (SetPrvStatus);
#endif

   /* status from sys$open() */
   if (VMSnok (status))
   {
      if (Debug) fprintf (stdout, "sys$open() %%X%08.08X\n", status);
      cptr = MapUrl_ReportProblem (rlptr, NULL, status);
      ErrorString[0] = '\0';
      strcpy (ErrorString+1, cptr);
      return (ErrorString);
   }

   rlptr->LoadFileName[MapFileNam.nam$b_rsl] = '\0';
   if (Debug) fprintf (stdout, "LoadFileName |%s|\n", rlptr->LoadFileName);

   /* get the configuration file revision date and time */
   memcpy (&rlptr->RevBinTime,
           &MapFileXabDat.xab$q_rdt,
           sizeof(rlptr->RevBinTime));

   /* record access block */
   MapFileRab = cc$rms_rab;
   MapFileRab.rab$l_fab = &MapFileFab;
   /* try and optimise sequential read (next 3 lines) */
   MapFileRab.rab$b_mbc = 6;
   MapFileRab.rab$b_mbf = 2;
   MapFileRab.rab$l_rop = RAB$M_RAH;
   MapFileRab.rab$l_ubf = rlptr->LinePtr = Line;
   MapFileRab.rab$w_usz = sizeof(Line)-1;

   if (VMSnok (status = sys$connect (&MapFileRab, 0, 0)))
   {
      sys$close (&MapFileFab, 0, 0);
      if (Debug) fprintf (stdout, "sys$connect() %%X%08.08X\n", status);
      cptr = MapUrl_ReportProblem (rlptr, NULL, status);
      ErrorString[0] = '\0';
      strcpy (ErrorString+1, cptr);
      return (ErrorString);
   }

   /*************************/
   /* read the mapping file */
   /*************************/

   rlptr->LineNumber = 0;

   while (VMSok (status = sys$get (&MapFileRab, 0, 0)))
   {
      rlptr->LineNumber++;
      Line[MapFileRab.rab$w_rsz] = '\0';
      if (Debug) fprintf (stdout, "Line |%s|\n", Line);

      cptr = Line;
      while (isspace(*cptr)) cptr++;

      /* if a blank or comment line, then ignore */
      if (!*cptr || *cptr == '!' || *cptr == '#') continue;

      /* get the first character, then buffer the rule in case of error */
      RuleChar = toupper(*cptr);
      zptr = (sptr = rlptr->RuleName) + sizeof(rlptr->RuleName);
      while (*cptr && !isspace(*cptr) && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr)
      {
         *--sptr = '\0';
         MapUrl_ReportProblem (rlptr, ProblemConfused, 0);
         continue;
      }
      *sptr = '\0';
      if (sptr > rlptr->RuleName)
         if (*(sptr-1) == '+')
            IsCgiPlusScript = true;
         else
            IsCgiPlusScript = false;
      while (isspace(*cptr)) cptr++;

      /* copy template into scratch area */
      zptr = (sptr = Scratch) + sizeof(Scratch);
      while (*cptr && !isspace(*cptr) && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr)
      {
         *--sptr = '\0';
         MapUrl_ReportProblem (rlptr, ProblemConfused, 0);
         continue;
      }
      *sptr++ = '\0';
      TemplateLength = sptr - Scratch;
      /* skip intervening white-space */
      while (isspace(*cptr)) cptr++;

      /* copy (optional) result into scratch area, */
      ResultPtr = sptr;
      while (*cptr && !isspace(*cptr) && sptr < zptr) *sptr++ = *cptr++;
      if (sptr >= zptr)
      {
         MapUrl_ReportProblem (rlptr, ProblemConfused, 0);
         continue;
      }
      *sptr++ = '\0';
      TotalLength = sptr - Scratch;

      if (Debug)
         fprintf (stdout, "|%c|%d|%s|%s|\n",
                  RuleChar, IsCgiPlusScript, Scratch, ResultPtr);

      /******************************/
      /* basic consistency checking */
      /******************************/

      TemplateWildcardCount = 0;
      for (cptr = Scratch; *cptr; cptr++)
         if (*cptr == '*') TemplateWildcardCount++;

      ResultWildcardCount = 0;
      for (cptr = ResultPtr; *cptr; cptr++)
         if (*cptr == '*') ResultWildcardCount++;

      RequiresResult = false;
      switch (RuleChar)
      {
         case 'E' :
         case 'S' :
            if (ResultWildcardCount != 1 && TemplateWildcardCount != 1)
            {
               MapUrl_ReportProblem (rlptr, ProblemWildcardMapping, 0);
               continue;
            }
            RequiresResult = true;
            break;

         case 'F' :
         case 'P' :
            break;

         case 'M' :
         case 'R' :
            RequiresResult = true;
            break;

         default :
            MapUrl_ReportProblem (rlptr, "?", 0);
            continue;
      }

      if (!Scratch[0])
      {
         /* there must be a "template" to map from */
         MapUrl_ReportProblem (rlptr, ProblemTemplateRequired, 0);
         continue;
      }

      if (RequiresResult && !ResultPtr[0])
      {
         /* there must be a "result" to map to */
         MapUrl_ReportProblem (rlptr, ProblemResultRequired, 0);
         continue;
      }

      if (ResultWildcardCount > TemplateWildcardCount)
      {
         /* cannot wildcard map more elements than in template */
         MapUrl_ReportProblem (rlptr, ProblemWildcardMapping, 0);
         continue;
      }

      /************************************/
      /* enter the rule into the database */
      /************************************/

      if ((mrptr = (struct MappingRuleStruct*)
           malloc (MappingRuleStructOverhead+TotalLength)) == NULL)
      {
         /* no point in continuing if we can't get memory for essentials */
#ifdef IP_MAPURL
         ErrorExitVmsStatus (vaxc$errno, "malloc()", __FILE__, __LINE__);
#else
         exit (vaxc$errno);
#endif
      }

      memcpy ((char*)mrptr+MappingRuleStructOverhead, Scratch, TotalLength);
      mrptr->NextPtr = NULL;
      mrptr->MapFileLineNumber = rlptr->LineNumber;
      mrptr->RuleType = RuleChar;
      mrptr->IsCgiPlusScript = IsCgiPlusScript;
      mrptr->TemplatePtr = (char*)mrptr + MappingRuleStructOverhead;
      mrptr->ResultPtr = (char*)mrptr + MappingRuleStructOverhead +
                         TemplateLength;
      if (rlptr->HeadPtr == NULL)
         rlptr->HeadPtr = mrptr;
      else
         rlptr->TailPtr->NextPtr = mrptr;
      rlptr->TailPtr = mrptr;
   }

   /**************************/
   /* close the mapping file */
   /**************************/

   /** if (Debug) fprintf (stdout, "sys$get() %%X%08.08X\n", status); **/
   if (status == RMS$_EOF) status = SS$_NORMAL;

   sys$close (&MapFileFab, 0, 0);

   if (VMSok (status))
      *(unsigned long*)ErrorString = 0;
   else
   {
      MapUrl_FreeRules (rlptr);
      cptr = MapUrl_ReportProblem (rlptr, NULL, status);
      ErrorString[0] = '\0';
      strcpy (ErrorString+1, cptr);
   }

   return (ErrorString);
}

/*****************************************************************************/
/*
This function formats an error report.  All lines are concatenated onto a
single string of dynamically allocated memory that (obviously) grows as
reports are added to it.  This string is then output if loading the server
configuration or is available for inclusion in an HTML page.
*/

char* MapUrl_ReportProblem
(
struct MappingRuleListStruct *rlptr,
char *Explanation,
int StatusValue
)
{
   static $DESCRIPTOR (RuleFaoDsc,
                       "%HTTPD-W-MAPURL, \"!AZ\" rule !AZ at line !UL\n");
   static $DESCRIPTOR (ExplanationFaoDsc,
                       "%HTTPD-W-MAPURL, !AZ at line !UL\n");

   static char  Buffer [256];   
   static $DESCRIPTOR (BufferDsc, Buffer);

   int  status;
   unsigned short  Length;
   char  HtmlBuffer [512];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MapUrl_ReportProblem()\n");

   rlptr->ProblemCount++;

   if (StatusValue)
   {
      if (VMSok (status = sys$getmsg (StatusValue, &Length, &BufferDsc, 0, 0)))
         Buffer[Length++] = '\n';
   }
   else
   if (Explanation == NULL)
      status = sys$fao (&RuleFaoDsc, &Length, &BufferDsc,
                        rlptr->RuleName, "problem", rlptr->LineNumber);
   else
   if (Explanation[0] == '?')
      status = sys$fao (&RuleFaoDsc, &Length, &BufferDsc,
                        rlptr->RuleName, "unknown", rlptr->LineNumber);
   else
      status = sys$fao (&ExplanationFaoDsc, &Length, &BufferDsc,
                        Explanation, rlptr->LineNumber);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
#ifdef IP_MAPURL
         ErrorExitVmsStatus (status, "sys$fao()", __FILE__, __LINE__);
#else
         exit (status);
#endif
   else
      Buffer[Length] = '\0';
   if (Debug) fprintf (stdout, "|%s|\n", Buffer);
   if (rlptr->RequestPtr == NULL) fputs (Buffer, stdout);

   if ((Length =
       MapUrl_CopyToHtml (HtmlBuffer, sizeof(HtmlBuffer), Buffer, -1)) < 0)
   {
#ifdef IP_MAPURL
      ErrorExitVmsStatus (SS$_BUFFEROVF, "CopyToHtml()", __FILE__, __LINE__);
#else
      exit (SS$_BUFFEROVF & 0xfffffffe);
#endif
   }

   if ((rlptr->ProblemReportPtr =
        realloc (rlptr->ProblemReportPtr, rlptr->ProblemReportLength+Length+1))
       == NULL)
#ifdef IP_MAPURL
         ErrorExitVmsStatus (vaxc$errno, "realloc()", __FILE__, __LINE__);
#else
         exit (status);
#endif
   /* include the terminating null */
   memcpy (rlptr->ProblemReportPtr+rlptr->ProblemReportLength,
           HtmlBuffer, Length+1);
   rlptr->ProblemReportLength += Length;

   return (Buffer);
}

/*****************************************************************************/
/*
Free all entries in the singly-linked rule list.
*/ 
 
MapUrl_FreeRules (struct MappingRuleListStruct *rlptr)

{
   register char  *cptr;
   register struct MappingRuleStruct  *mrptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MapUrl_FreeRules()\n");

   mrptr = rlptr->HeadPtr;
   while (mrptr != NULL)
   {
      cptr = (char*)mrptr;
      mrptr = mrptr->NextPtr;
      free (cptr);
   }
   rlptr->HeadPtr = rlptr->TailPtr = NULL;
}

/*****************************************************************************/
/*
Convert a URL-style specification into a RMS-style specification.  For 
example: "/disk/dir1/dir2/file.txt" into "disk:[dir1.dir2]file.txt".  Non-RMS 
characters (non-alphanumberic and non-"$_-" are converted to "$'.  Where a 
specification has multiple ".", all but the final one of the file component is 
converted to a "$".
*/ 
 
MapUrl_UrlToVms
(
char *PathPtr,
char *VmsPtr
)
{
   register int  ncnt, ccnt;
   register char  *cptr, *pptr, *sptr;

   char  PathComponents [256];

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MapUrl_UrlToVms() |%s|\n", PathPtr);

   if (!PathPtr[0])
   {
      VmsPtr[0] = VmsPtr[1] = '\0';
      return;
   }

   /******************************************/
   /* parse out each "/" delimited component */
   /******************************************/

   /* each component becomes a null-terminated string in a character array */
   sptr = PathComponents;
   pptr = PathPtr;
   ccnt = 0;
   while (*pptr)
   {
      if (*pptr == '/') pptr++;
      /* copy, converting all non-RMS characters into exclamation marks */
      cptr = sptr;
      while (*pptr && *pptr != '/')
      {
         if (isalnum (*pptr))
            *sptr++ = toupper(*pptr++);
         else
         if (pptr[0] == '.' && pptr[1] == '.' && pptr[2] == '.')
         {
            *sptr++ = *pptr++;
            *sptr++ = *pptr++;
            *sptr++ = *pptr++;
         }
         else
         if (*pptr == '$' || *pptr == '_' || *pptr == '-' ||
             *pptr == '*' || *pptr == '%' || *pptr == ';')
            *sptr++ = *pptr++;
         else
         {
            *sptr++ = '!';
            pptr++;
         }
      }
      *sptr++ = '\0';
      ccnt++;
      if (Debug) fprintf (stdout, "%d |%s|\n", ccnt, cptr);
   }

   /********************************/
   /* build VMS file specification */
   /********************************/

   if (ccnt == 1)
   {
      /* single component */
      cptr = PathComponents;
      sptr = VmsPtr;
      while (*cptr) *sptr++ = *cptr++;
      *sptr = '\0';
   }
   else
   if (ccnt == 2)
   {
      /***********************/
      /* top-level directory */
      /***********************/

      cptr = PathComponents;
      sptr = VmsPtr;
      /* start with the first component, the "device" */
      while (*cptr) *sptr++ = *cptr++;
      cptr++;
      /* append a Master File Directory component */
      strcpy (sptr, ":[000000]");
      while (*sptr) sptr++;
      /* add the second component, the "file" */
      while (*cptr) *sptr++ = *cptr++;
      *sptr = '\0';
   }
   else
   if (ccnt >= 3)
   {
      /****************/
      /* subdirectory */
      /****************/

      cptr = PathComponents;
      sptr = VmsPtr;
      /* start with the first component, the "device" */
      while (*cptr) *sptr++ = *cptr++;
      cptr++;
      *sptr++ = ':';
      *sptr++ = '[';
      if (cptr[0] == '.' && cptr[1] == '.' && cptr[2] == '.')
      {
         strcpy (sptr, "000000");
         sptr += 6;
      }
      for (ncnt = 2; ncnt < ccnt; ncnt++)
      {
         /* add the next component, a "directory" */
         if (cptr[0] == '.' && cptr[1] == '.' && cptr[2] == '.')
            while (*cptr) *sptr++ = *cptr++;
         else
         {
            if (ncnt > 2) *sptr++ = '.';
            while (*cptr) *sptr++ = *cptr++;
         }
         cptr++;
      }
      *sptr++ = ']';
      /* add the last component, the "file" */
      while (*cptr) *sptr++ = *cptr++;
      *sptr = '\0';
   }

   /*******************************************/
   /* convert intermediate non-RMS characters */
   /*******************************************/

   for (sptr = VmsPtr; *sptr && *sptr != ']'; sptr++)
      if (*sptr == '!') *sptr = '$';
   if (*sptr)
   {
      cptr = "";
      for (sptr++; *sptr; sptr++)
         if (*sptr == '!') *(cptr = sptr) = '$';
      if (*cptr) *cptr = '.';
   }

   if (!VmsPtr[0]) VmsPtr[1] = '\0';

   if (Debug) fprintf (stdout, "VmsPtr |%s|\n", VmsPtr);
}

/*****************************************************************************/
/*
Convert a VMS file specification into a URL-style specification.  For example:
"DEVICE:[DIR1.DIR2]FILE.TXT" into "/device/dir1/dir2/file.txt".
*/ 
 
MapUrl_VmsToUrl
(
char *PathPtr,
char *VmsPtr
)
{
   register char  *pptr, *vptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MapUrl_VmsToUrl() |%s|\n", VmsPtr);

   vptr = VmsPtr;
   pptr = PathPtr;
   *pptr++ = '/';
   /* copy the device and directory components */
   while (*vptr)
   {
      if (*vptr == ':' && *(vptr+1) == '[')
      {
         vptr++;
         vptr++;
         /* remove any reference to a Master File Directory */
         if (strncmp (vptr, "000000", 6) == 0)
            vptr += 6;
         else
            *pptr++ = '/';
      }
      if (*vptr == '.')
      {
         if (vptr[1] == '.' && vptr[2] == '.')
         {
            *pptr++ = '/';
            *pptr++ = *vptr++;
            *pptr++ = *vptr++;
            *pptr++ = *vptr++;
         }
         else
         {
            vptr++;
            *pptr++ = '/';
         }
      }
      else
      if (*vptr == ']')
      {
         vptr++;
         *pptr++ = '/';
         break;
      }
      else
         *pptr++ = tolower(*vptr++);
   }
   /* copy the file component */
   while (*vptr) *pptr++ = tolower(*vptr++);
   *pptr++ = '\0';
   *pptr = '\0';

   if (Debug) fprintf (stdout, "PathPtr |%s|\n", PathPtr);
}

/*****************************************************************************/
/*
Given the current document's URL path and the specified document's URL path, 
generate a resultant URL path based on whether the specified document's path 
is absolute or relative the the current document's path.  Insufficient space to
contain the result path will not crash the function and the result will be set
to an empty string.
*/ 

int MapUrl_VirtualPath
(
char *CurrentPath,
char *DocumentPath,
char *ResultPath,
int SizeOfResultPath
)
{
   register char  *cptr, *dptr, *sptr, *tptr, *zptr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "MapUrl_VirtualPath() |%s|%s|\n",
               CurrentPath, DocumentPath);

   if (*(dptr = DocumentPath) == '/' || strstr (DocumentPath, "//") != NULL) 
   {
      /*****************/
      /* absolute path */
      /*****************/

      strcpy (ResultPath, dptr);
      zptr = (sptr = ResultPath) + SizeOfResultPath;
      while (*dptr && sptr < zptr) *sptr++ = *dptr++;
      if (sptr >= zptr) return (-1);
      *sptr = '\0';
      if (Debug) fprintf (stdout, "ResultPath |%s|\n", ResultPath);
      return (sptr - ResultPath);
   }

   /* step over any "relative to this directory" syntax ("./") */
   if (dptr[0] == '.' && dptr[1] == '/') dptr += 2;
   if (*dptr != '.')
   {
      /*****************/
      /* inferior path */
      /*****************/

      zptr = (sptr = tptr = ResultPath) + SizeOfResultPath;
      cptr = CurrentPath;
      while (*cptr && sptr < zptr)
      {
         if (*cptr == '/') tptr = sptr+1;
         *sptr++ = *cptr++;
      }
      sptr = tptr;
      while (*dptr && sptr < zptr) *sptr++ = *dptr++;
      if (sptr >= zptr) return (-1);
      *sptr = '\0';
      if (Debug) fprintf (stdout, "ResultPath |%s|\n", ResultPath);
      return (sptr - ResultPath);
   }

   /*****************/
   /* relative path */
   /*****************/

   zptr = (sptr = tptr = ResultPath) + SizeOfResultPath;
   cptr = CurrentPath;
   while (*cptr && sptr < zptr)
   {
      if (*cptr == '/') tptr = sptr;
      *sptr++ = *cptr++;
   }
   sptr = tptr;
   /* loop, stepping back one level for each "../" in the document path */
   while (dptr[0] == '.' && dptr[1] == '.' && dptr[2] == '/')
   {
      dptr += 3;
      if (sptr > ResultPath) sptr--;
      while (sptr > ResultPath && *sptr != '/') sptr--;
   }
   if (sptr > ResultPath)
      sptr++;
   else
      *sptr++ = '/';
   while (*dptr && sptr < zptr) *sptr++ = *dptr++;
   if (sptr >= zptr) return (-1);
   *sptr = '\0';
   if (Debug) fprintf (stdout, "ResultPath |%s|\n", ResultPath);
   return (sptr - ResultPath);
}

/*****************************************************************************/
/*
Copy text from one string to another, converting characters forbidden to 
appear as plain-text in HTML.  For example the '<', '&', etc.  Convert these 
to the corresponding HTML character entities.  Returns the number of characters
added to the buffer or -1 to indicate buffer overflow.  Copies a maximum of
'Count' characters, or all id it is -1.
*/ 

int MapUrl_CopyToHtml
( 
char *Destination,
int SizeOfDestination,
char *Source,
int Count
)
{
    register char  *cptr, *sptr, *zptr;
    register int  ccnt;

    cptr = Source;
    zptr = (sptr = Destination) + SizeOfDestination;
    ccnt = Count;
    while (--ccnt && *cptr)
    {
       switch (*cptr)
       {
          case '<' :
             if (sptr+4 >= zptr) return (-1);
             memcpy (sptr, "&lt;", 4); sptr += 4; cptr++; break;
          case '>' :
             if (sptr+4 >= zptr) return (-1);
             memcpy (sptr, "&gt;", 4); sptr += 4; cptr++; break;
          case '&' :
             if (sptr+5 >= zptr) return (-1);
             memcpy (sptr, "&amp;", 5); sptr += 5; cptr++; break;
          case '\"' :
             if (sptr+6 >= zptr) return (-1);
             memcpy (sptr, "&quot;", 6); sptr += 6; cptr++; break;
          default :
             if (sptr >= zptr) return (-1);
             *sptr++ = *cptr++;
       }
    }
    if (sptr >= zptr) return (-1);
    *sptr = '\0';
    return (sptr-Destination);
}

#ifdef IP_MAPURL

/*****************************************************************************/
/*
*/

MapUrl_Report
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
boolean UseServerDatabase
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ:!UL ... Mapping Rules</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>HTTPd !AZ:!UL</H1>\n\
<H2>Mapping Rules</H2>\n\
!AZ\n");

   static $DESCRIPTOR (ProblemReportFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH><FONT COLOR=\"#ff0000\">!UL Problem!AZ !AZ</FONT></TH></TR>\n\
<TR><TD><PRE>!AZ</PRE></TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (CheckPathFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Check Path Against Rules</TH></TR>\n\
<TR><TD>\n\
<FORM METHOD=GET ACTION=\"!AZ\">\n\
<NOBR>\n\
Using <INPUT TYPE=radio NAME=server VALUE=yes CHECKED>server or \
<INPUT TYPE=radio NAME=server VALUE=no>file rules ...\n\
<BR><INPUT TYPE=text NAME=path SIZE=50>\n\
<INPUT TYPE=submit VALUE=\" check \">\n\
<INPUT TYPE=reset VALUE=\" reset \">\n\
</NOBR>\n\
</FORM>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>Source: &quot;!AZ&quot;</TH></TR>\n\
<TR><TD>\n\
<TABLE CELLPADDING=2 NOBORDER>\n\
<TR><TH ALIGN=RIGHT>File:</TH>\
<TD ALIGN=LEFT>!AZ</TD>\
<TD>&nbsp;</TD><TH>[<A HREF=\"!AZ\">View</A>]</TH>\
</TR>\n\
<TR><TH ALIGN=RIGHT>!AZ</TH>\
<TD ALIGN=LEFT>!AZ</TD></TR>\n\
</TABLE>\n\
</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>rule</TH><TH>path</TH><TH>template</TH></TR>\n");

   static $DESCRIPTOR (RuleFaoDsc,
"<TR><TH>!AZ</TH><TD>!AZ</TD><TD>!AZ</TD></TR>\n");

   static $DESCRIPTOR (EndPageFaoDsc,
"!AZ\
</TABLE>\n\
</BODY>\n\
</HTML>\n");

   register unsigned long  *vecptr;
   register char  *cptr;
   register struct MappingRuleStruct  *mrptr;

   boolean  CheckFromFile;
   int  status,
        EntryCount;
   unsigned long  FaoVector [32];
   unsigned short  Length;
   char  Buffer [2048];
   $DESCRIPTOR (BufferDsc, Buffer);
   struct MappingRuleListStruct  RulesList;
   struct MappingRuleListStruct  *rlptr;
   struct MappingRuleListStruct  *GlobalMappingRulesListPtrBuffer;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MapUrl_ReportRuleCheck()\n");

   if (UseServerDatabase)
   {
      /* use mapping rules from supplied pointer */
      rlptr = GlobalMappingRulesListPtr;
   }
   else
   {
      /* load temporary set of rules from mapping file */
      memset (rlptr = &RulesList, 0, sizeof(struct MappingRuleListStruct)); 
      rlptr->RequestPtr = rqptr;
      cptr = MapUrl_LoadRules (rlptr);
      /* if error reported */
      if (!cptr[0] && cptr[1])
      {
         rqptr->ResponseStatusCode = 403;
         ErrorGeneral (rqptr, cptr+1, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }
   }

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
   {
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;

   *vecptr++ = HtmlSgmlDoctype;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   QioNetWrite (rqptr, 0, Buffer, Length);

   if (rlptr->ProblemReportLength)
   {
      vecptr = FaoVector;
      if ((*vecptr++ = rlptr->ProblemCount) == 1)
         *vecptr++ = "";
      else
         *vecptr++ = "s";
      if (UseServerDatabase)
         *vecptr++ = "Reported At Startup";
      else
         *vecptr++ = "Detected During Load";
      *vecptr++ = rlptr->ProblemReportPtr;

      status = sys$faol (&ProblemReportFaoDsc, &Length, &BufferDsc,
                         &FaoVector);

      /* if using a local database free report memory now! */
      if (!UseServerDatabase) free (rlptr->ProblemReportPtr);

      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }

      Buffer[Length] = '\0';
      QioNetWrite (rqptr, 0, Buffer, Length);
   }

   vecptr = FaoVector;

   *vecptr++ = HttpdInternalReportRules;
   if (UseServerDatabase)
      *vecptr++ = "Server";
   else
      *vecptr++ = "File";
   *vecptr++ = rlptr->LoadFileName;
   *vecptr++ = MapUrl (NULL, rlptr->LoadFileName, NULL, NULL);
   if (UseServerDatabase)
   {
      *vecptr++ = "Loaded:";
      *vecptr++ = DayDateTime (&rlptr->LoadBinTime, 20);
   }
   else
   {
      *vecptr++ = "Revised:";
      *vecptr++ = DayDateTime (&rlptr->RevBinTime, 20);
   }

   status = sys$faol (&CheckPathFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   QioNetWrite (rqptr, 0, Buffer, Length);

   status = SS$_NORMAL;
   EntryCount = 0;

   for (mrptr = rlptr->HeadPtr; mrptr != NULL; mrptr = mrptr->NextPtr)
   {
      if (Debug)
         fprintf (stdout, "Rule |%c|%s|%s|\n",
                  mrptr->RuleType, mrptr->TemplatePtr, mrptr->ResultPtr);

      EntryCount++;

      if (mrptr->RuleType == 'E')
         if (mrptr->IsCgiPlusScript)
            cptr = "exec+";
         else
            cptr = "exec";
      else
      if (mrptr->RuleType == 'F')
         cptr = "fail";
      else
      if (mrptr->RuleType == 'M')
         cptr = "map";
      else
      if (mrptr->RuleType == 'P')
         cptr = "pass";
      else
      if (mrptr->RuleType == 'R')
         cptr = "redirect";
      else
      if (mrptr->RuleType == 'S')
         if (mrptr->IsCgiPlusScript)
            cptr = "script+";
         else
            cptr = "script";
      else
         cptr = "*ERROR*";

      vecptr = FaoVector;
      *vecptr++ = cptr;
      *vecptr++ = mrptr->TemplatePtr;
      *vecptr++ = mrptr->ResultPtr;
      status = sys$faol (&RuleFaoDsc, &Length, &BufferDsc, &FaoVector);
      if (VMSnok (status)) break;

      Buffer[Length] = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", Buffer);
      QioNetWrite (rqptr, 0, Buffer, Length);
   }

   if (!UseServerDatabase) MapUrl_FreeRules (&RulesList);

   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;
   if (EntryCount)
      *vecptr++ = "";
   else
      *vecptr++ = "<TR><TD COLSPAN=3><I>(none)</I></TD></TR>\n";

   status = sys$faol (&EndPageFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   QioNetWrite (rqptr, 0, Buffer, Length);

   if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);

   rqptr->ResponseStatusCode = 200;
}

/*****************************************************************************/
/*
Display mapping rules used to resolve the path in the query string.  Called
from the Request.c modules.  For HTTP report, uses blocking network write.
*/

MapUrl_ReportRuleCheck
(
struct RequestStruct *rqptr,
void *NextTaskFunction,
char *Path,
boolean UseServerDatabase
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>HTTPd !AZ:!UL ... Mapping Rule Check</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>HTTPd !AZ:!UL</H1>\n\
<H2>Mapping Rules Check</H2>\n\
!AZ\n");

   static $DESCRIPTOR (ProblemReportFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH><FONT COLOR=\"#ff0000\">!UL Problem!AZ !AZ</FONT></TH></TR>\n\
<TR><TD><PRE>!AZ</PRE></TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (SourceFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH COLSPAN=2>Source: &quot;!AZ&quot;</TH></TR>\n\
<TR><TH>File</TH><TD>!AZ</TD></TR>\n\
<TR><TH>!AZ</TH><TD>!AZ</TD></TR>\n\
</TABLE>\n\
\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH>path</TH><TH>match</TH><TH COLSPAN=3>rule</TH></TR>\n\
<TR><TH></TH><TH></TH><TH>rule</TH><TH>path</TH><TH>template</TH></TR>\n");

   static $DESCRIPTOR (EndErrorFaoDsc,
"</TABLE>\n\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TD>!AZ</TD></TR>\n\
</TABLE>\n\
</BODY>\n\
</HTML>\n");

   static $DESCRIPTOR (EndPageFaoDsc,
"</TABLE>\n\
<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH></TH><TH>path</TH><TH>VMS</TH></TR>\n\
<TR><TH>path</TH><TD>!AZ</TD><TD>!AZ</TD></TR>\n\
<TR><TH>script!AZ</TH><TD>!AZ</TD><TD>!AZ</TD></TR>\n\
</TABLE>\n\
</BODY>\n\
</HTML>\n");

   register unsigned long  *vecptr;
   register char  *cptr;

   int  status;
   unsigned long  FaoVector [32];
   unsigned short  Length;
   char  Buffer [1024],
         Script [256],
         ScriptVms [256],
         Vms [256];
   $DESCRIPTOR (BufferDsc, Buffer);
   struct MappingRuleListStruct  RulesList;
   struct MappingRuleListStruct  *rlptr;
   struct MappingRuleListStruct  *GlobalMappingRulesListPtrBuffer;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MapUrl_ReportRuleCheck()\n");

   if (UseServerDatabase)
   {
      /* use mapping rules from supplied pointer */
      memcpy (rlptr = &RulesList,
              GlobalMappingRulesListPtr,
              sizeof(struct MappingRuleListStruct)); 
   }
   else
   {
      /* load temporary set of rules from mapping file */
      memset (rlptr = &RulesList, 0, sizeof(struct MappingRuleListStruct)); 
      rlptr->RequestPtr = rqptr;
      cptr = MapUrl_LoadRules (rlptr);
      /* if error reported */
      if (!cptr[0] && cptr[1])
      {
         rqptr->ResponseStatusCode = 403;
         ErrorGeneral (rqptr, cptr+1, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }
   }

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
   {
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;

   *vecptr++ = HtmlSgmlDoctype;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;
   *vecptr++ = DayDateTime (&rqptr->BinaryTime, 20);

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   QioNetWrite (rqptr, 0, Buffer, Length);

   if (rlptr->ProblemReportLength)
   {
      vecptr = FaoVector;
      if ((*vecptr++ = rlptr->ProblemCount) == 1)
         *vecptr++ = "";
      else
         *vecptr++ = "s";
      if (UseServerDatabase)
         *vecptr++ = "Reported At Startup";
      else
         *vecptr++ = "Detected During Load";
      *vecptr++ = rlptr->ProblemReportPtr;

      status = sys$faol (&ProblemReportFaoDsc, &Length, &BufferDsc,
                         &FaoVector);

      /* if using a local database free report memory now! */
      if (!UseServerDatabase) free (rlptr->ProblemReportPtr);

      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }

      Buffer[Length] = '\0';
      QioNetWrite (rqptr, 0, Buffer, Length);
   }

   vecptr = FaoVector;

   if (UseServerDatabase)
      *vecptr++ = "Server";
   else
      *vecptr++ = "File";
   *vecptr++ = rlptr->LoadFileName;
   if (UseServerDatabase)
   {
      *vecptr++ = "Loaded:";
      *vecptr++ = DayDateTime (&rlptr->LoadBinTime, 20);
   }
   else
   {
      *vecptr++ = "Revised:";
      *vecptr++ = DayDateTime (&rlptr->RevBinTime, 20);
   }

   status = sys$faol (&SourceFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   QioNetWrite (rqptr, 0, Buffer, Length);

   rlptr->RequestPtr = rqptr;
   GlobalMappingRulesListPtrBuffer = GlobalMappingRulesListPtr;
   GlobalMappingRulesListPtr = &RulesList;

   Script[0] = ScriptVms[0] = Vms[0] = '\0';
   cptr = MapUrl (Path, Vms, Script, ScriptVms);

   GlobalMappingRulesListPtr = GlobalMappingRulesListPtrBuffer;

   if (!UseServerDatabase) MapUrl_FreeRules (rlptr);

   if (!cptr[0] && cptr[1])
      status = sys$fao (&EndErrorFaoDsc, &Length, &BufferDsc,
                        cptr+1);
   else
   {
      if (Script[0] == '+')
      {
         Script[0] = '/';
         status = sys$fao (&EndPageFaoDsc, &Length, &BufferDsc,
                           cptr, Vms, " (CGIplus)", Script, ScriptVms);
      }
      else
         status = sys$fao (&EndPageFaoDsc, &Length, &BufferDsc,
                           cptr, Vms, "", Script, ScriptVms);
   }

   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$fao()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   Buffer[Length] = '\0';
   QioNetWrite (rqptr, 0, Buffer, Length);

   if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);

   rqptr->ResponseStatusCode = 200;
}

/*****************************************************************************/
/*
*/

MapUrl_ReportRule
(
register struct MappingRuleStruct  *mrptr,
char *PathPtr,
boolean Match
)
{
   static $DESCRIPTOR (HtmlRuleFaoDsc,
"<TR><TD>!AZ</TD><TH ALIGN=CENTER>!AZ</TH>\
<TH>!AZ</TH><TD>!AZ</TD><TD>!AZ</TD></TR>\n");

   int  status;
   unsigned short  Length;
   char  Buffer [512];
   char  *MatchPtr,
         *RulePtr;
   $DESCRIPTOR (BufferDsc, Buffer);

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MapUrl_ReportRule()\n");

   switch (mrptr->RuleType)
   {
      case 'E' :
         if (mrptr->IsCgiPlusScript)
            { RulePtr = "exec+"; break; }
         else
            { RulePtr = "exec"; break; }
      case 'F' :
         RulePtr = "fail"; break;
      case 'M' :
         RulePtr = "map"; break;
      case 'P' :
         RulePtr = "pass"; break;
      case 'R' :
         RulePtr = "redirect"; break;
      case 'S' :
         if (mrptr->IsCgiPlusScript)
            { RulePtr = "script+"; break; }
         else
            { RulePtr = "script"; break; }
      default :
         RulePtr = "*ERROR*";
   }

   if (Match)
      MatchPtr = "X";
   else
      MatchPtr = "";

   status = sys$fao (&HtmlRuleFaoDsc, &Length, &BufferDsc,
                     PathPtr, MatchPtr,
                     RulePtr, mrptr->TemplatePtr, mrptr->ResultPtr);
   if (VMSok (status) && status != SS$_BUFFEROVF)
   {
      Buffer[Length] = '\0';
      if (Debug) fprintf (stdout, "|%s|\n", Buffer);
      QioNetWrite (GlobalMappingRulesListPtr->RequestPtr, 0, Buffer, Length);
      return;
   }
   QioNetWrite (GlobalMappingRulesListPtr->RequestPtr, 0, "FAO() ERROR", 11);
}

/*****************************************************************************/
/*
Display mapping rules used to resolve the path in the query string.  Called
from the Request.c modules.  For HTTP report, uses blocking network write.
*/

MapUrl_ControlReload
(
struct RequestStruct *rqptr,
void *NextTaskFunction
)
{
   static $DESCRIPTOR (ResponseFaoDsc,
"!AZ\
<HTML>\n\
<HEAD>\n\
!AZ\
<TITLE>Success 200</TITLE>\n\
</HEAD>\n\
<BODY>\n\
<H1>SUCCESS!!</H1>\n\
<P>Reported by server.\n\
<P>Server !AZ:!UL mapping rules reloaded.\n");

   static $DESCRIPTOR (ProblemReportFaoDsc,
"<P><TABLE CELLPADDING=2 BORDER>\n\
<TR><TH><FONT COLOR=\"#ff0000\">!UL Problem!AZ \
Detected During Load</FONT></TH></TR>\n\
<TR><TD><PRE>!AZ</PRE></TD></TR>\n\
</TABLE>\n");

   static $DESCRIPTOR (EndPageFaoDsc,
"</BODY>\n\
</HTML>\n");

   register char  *cptr;
   register unsigned long  *vecptr;

   int  status;
   unsigned long  FaoVector [16];
   unsigned short  Length;
   char  Buffer [4096];
   $DESCRIPTOR (BufferDsc, Buffer);
   struct MappingRuleListStruct  *rlptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "MapUrl_ControlReload()\n");

   rlptr = GlobalMappingRulesListPtr;
   cptr = MapUrl_LoadRules (rlptr);
   /* if error reported */
   if (!cptr[0] && cptr[1])
   {
      rqptr->ResponseStatusCode = 403;
      ErrorGeneral (rqptr, cptr+1, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   rqptr->ResponsePreExpired = PRE_EXPIRE_ADMIN;
   if ((rqptr->ResponseHeaderPtr = HttpHeader200Html (rqptr)) == NULL)
   {
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   vecptr = FaoVector;

   *vecptr++ = HtmlSgmlDoctype;
   *vecptr++ = HtmlMetaInfo (rqptr, NULL);

   *vecptr++ = ServerHostName;
   *vecptr++ = ServerPort;

   status = sys$faol (&ResponseFaoDsc, &Length, &BufferDsc, &FaoVector);
   if (VMSnok (status) || status == SS$_BUFFEROVF)
   {
      rqptr->ErrorTextPtr = "sys$faol()";
      ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
      return;
   }

   QioNetWrite (rqptr, 0, Buffer, Length);

   if (rlptr->ProblemReportLength)
   {
      vecptr = FaoVector;
      if ((*vecptr++ = rlptr->ProblemCount) == 1)
         *vecptr++ = "";
      else
         *vecptr++ = "s";
      *vecptr++ = rlptr->ProblemReportPtr;

      status = sys$faol (&ProblemReportFaoDsc, &Length, &BufferDsc,
                         &FaoVector);

      if (VMSnok (status) || status == SS$_BUFFEROVF)
      {
         rqptr->ErrorTextPtr = "sys$faol()";
         ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
         if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         return;
      }

      Buffer[Length] = '\0';
      QioNetWrite (rqptr, 0, Buffer, Length);
   }

   QioNetWrite (rqptr, 0, EndPageFaoDsc.dsc$a_pointer,
                          EndPageFaoDsc.dsc$w_length);

   if (VMSnok (status = sys$dclast (NextTaskFunction, rqptr, 0)))
      ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);

   rqptr->ResponseStatusCode = 200;
}

/*****************************************************************************/

#endif /* IP_MAPURL */

