/*****************************************************************************/
/*
                                   Net.c

The networking essentials are based on DEC TCP/IP Services for OpenVMS.  The 
QIO interface was obviously chosen because of its asynchronous, non-blocking 
capabilities.  Although some functionalilty could have been implemented using 
the BSD sockets abstraction and only the asynchronous I/O done using the QIO I 
opted for working out the QIOs.  It wasn't too difficult and avoids the 
(slight) extra overhead of the sockets layer.  The TGV MultiNet port follows 
the basic outline of the DEC TCP/IP functionality.  With MultiNet I couldn't 
find a way of implementing the gethostbyaddr() functionality using QIOs, as 
with DEC TCP/IP.  A slightly less elegant kludge gets around that.  I 
deliberately differentiated the DEC and TGV interfaces, even when there was 
some functional overlap (e.g. IO$_WRITEVBLK and IO$_SEND).  It wasn't much 
extra work and helps keep obvious the fact of the two APIs. 

With resource wait explicitly enabled all sys$qio()s should wait until
resources become available.  The only execption is ASTLM, a fundamental
requirement for this server.  If SS$_EXQUOTA is returned from a sys$qio()
(i.e. without wait for completion, and therefore depending on AST delivery)
then EXIT THE SERVER with error status!

It appears as if Multinet (3.4 VAX at least) can return zero bytes and a
normal status upon client disconnection.  The was observed in RequestGet()
when a client aborted the connection before sending any data.  This caused an
endless AST loop reading zero bytes, normal status and never filling the
request buffer.  Multinet is now checked for a return of zero bytes and the
status is changed to an error if returned normal.


VERSION HISTORY
---------------
01-FEB-97  MGD  HTTPd version 4
27-SEP-96  MGD  add dummy read for early error reporting
12-APR-96  MGD  observed Multinet disconnection/zero-byte behaviour
                (request now aborts if Multinet returns zero bytes)
03-JAN-96  MGD  support for both DEC TCP/IP Services and TGV MultiNet
01-DEC-95  MGD  NetBufferOutput() for improving network IO;
20-DEC-94  MGD  multi-threaded version
20-JUN-94  MGD  single-threaded version
*/
/*****************************************************************************/

/* standard C header files */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

/* VMS related header files */
#include <descrip.h>
#include <dvidef.h>
#include <iodef.h>
#include <prvdef.h>
#include <psldef.h>
#include <ssdef.h>
#include <stsdef.h>

/* Internet-related header files */
#ifdef IP_UCX
#   include <socket.h>
#   include <in.h>
#   include <netdb.h>
#   include <inet.h>
#   include <ucx$inetdef.h>
#endif

#ifdef IP_MULTINET
#   include "multinet_root:[multinet.include.sys]types.h"
#   include "multinet_root:[multinet.include.sys]socket.h"
#   include "multinet_root:[multinet.include.vms]inetiodef.h"
#   include "multinet_root:[multinet.include.netinet]in.h"
#   include "multinet_root:[multinet.include]netdb.h"
#endif

/* application-related header files */
#include "wasd.h"
#include "net.h"
#include "error.h"

/******************/
/* global storage */
/******************/

int  QueueLength = 5,
     ServerHostNameLength;

unsigned long  ClientInternetAddress32bit;

unsigned short  ClientChannel,
                ServerChannel;

char  ClientHostName [128],
      ClientInternetAddress [32],
      ServerHostName [128],
      ServerInternetAddress [32];

#ifdef IP_UCX
   $DESCRIPTOR (InetDeviceDsc, "UCX$DEVICE");
#endif

#ifdef IP_MULTINET
   $DESCRIPTOR (InetDeviceDsc, "INET0:");
#endif

struct AnIOsb  ServerIOsb;

struct sockaddr_in  ServerSocketName,
                    ClientSocketName;

struct hostent  *HostEntryPtr;

int  OptionEnabled = 1;

#ifdef IP_UCX

struct {
   unsigned int  Length;
   char  *Address;
} ServerSocketNameItem =
   { sizeof(ServerSocketName), &ServerSocketName };

int  ClientSocketNameLength;
struct {
   unsigned int  Length;
   char  *Address;
   unsigned int  *LengthPtr;
} ClientSocketNameItem =
   { sizeof(ClientSocketName), &ClientSocketName, &ClientSocketNameLength };

struct {
   int  l_onoff;
   int  l_linger;
} LingerSeconds = { 1, 10 };

struct {
   unsigned short  Length;
   unsigned short  Parameter;
   char  *Address;
} ReuseAddress =
   { sizeof(OptionEnabled), UCX$C_REUSEADDR, &OptionEnabled },
  ReuseAddressSocketOption =
   { sizeof(ReuseAddress), UCX$C_SOCKOPT, &ReuseAddress },
  LingerOnClose =
   { sizeof(LingerSeconds), UCX$C_LINGER, &LingerSeconds },
  LingerOnCloseOption =
   { sizeof(LingerOnClose), UCX$C_SOCKOPT, &LingerOnClose };

unsigned short  TcpSocket [2] =
   { UCX$C_TCP, INET_PROTYP$C_STREAM };

#endif /* IP_UCX */

/********************/
/* external storage */
/********************/

extern boolean  Debug;
extern boolean  CliLoggingDisabled;
extern boolean  CliLoggingEnabled;
extern boolean  CliMonitorDisabled;
extern boolean  CliMonitorEnabled;
extern boolean  ControlExitRequested;
extern boolean  ControlRestartRequested;
extern boolean  LoggingEnabled;
extern int  CliServerPort;
extern int  CurrentConnectCount;
extern int  MaxConcurrentConnections;
extern int  OutputBufferSize;
extern int  ServerPort;
extern char  CliLogFileName[];
extern char  ControlBuffer[];
extern char  ErrorSanityCheck[];
extern char  *AcceptHostsPtr;
extern char  *RejectHostsPtr;
extern struct AccountingStruct  Accounting;
extern struct ConfigStruct  Config;

/*****************************************************************************/
/*
Create a socket for the server to listen at.  Then in a infinite loop accept 
connections from clients.  The connection request is handled at normal 
execution level (not AST delivery level).  User-Mode AST delivery is disabled 
by an AST function called when a connection request is received.  This is done 
to ensure processing of the new connection occurs (relatively) uninterrupted.  
It is performed at normal processing level to allow host name processing to 
occur if necessary (some host database routines will not execute at AST-
delivery level).
*/ 

NetListen ()

{
   static unsigned long  SysPrvMask [2] = { PRV$M_SYSPRV, 0 };

   int  status;
   unsigned short  Length;
   struct RequestStruct  *rqptr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetListen()\n");

   /*******************************/
   /* set up the listening socket */
   /*******************************/

   /* assign a channel to the internet template device */
   if (VMSnok (status = sys$assign (&InetDeviceDsc, &ServerChannel, 0, 0)))
      ErrorExitVmsStatus (status, "assign server channel", __FILE__, __LINE__);

   /* make the channel a TCP, connection-oriented socket */

   /* need SYSPRV to bind to a well-known port (mark as privileged) */
   if (VMSnok (status = sys$setprv (1, &SysPrvMask, 0, 0)))
      ErrorExitVmsStatus (status, "enabling SYSPRV", __FILE__, __LINE__);

#ifdef IP_UCX
   if (Debug) fprintf (stdout, "sys$qio() IO$_SETMODE\n");
   status = sys$qiow (0, ServerChannel, IO$_SETMODE, &ServerIOsb, 0, 0,
                      &TcpSocket, 0, 0, 0, &ReuseAddressSocketOption, 0);

   if (Debug)
      fprintf (stdout, "sys$qiow() %%X%08.08X IOsb.Status %%X%08.08X\n",
               status, ServerIOsb.Status);
   if (VMSok (status) && VMSnok (ServerIOsb.Status))
      status = ServerIOsb.Status;
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "create server socket", __FILE__, __LINE__);
#endif

#ifdef IP_MULTINET
   if (Debug) fprintf (stdout, "sys$qio() IO$_SOCKET\n");
   status = sys$qiow (0, ServerChannel, IO$_SOCKET, &ServerIOsb, 0, 0,
                      AF_INET, SOCK_STREAM, 0, 0, 0, 0);

   if (Debug)
      fprintf (stdout, "sys$qiow() %%X%08.08X IOsb.Status %%X%08.08X\n",
               status, ServerIOsb.Status);
   if (VMSok (status) && VMSnok (ServerIOsb.Status))
      status = ServerIOsb.Status;
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "create server socket", __FILE__, __LINE__);

   if (Debug) fprintf (stdout, "sys$qio() IO$_SETSOCKOPT\n");
   status = sys$qiow (0, ServerChannel, IO$_SETSOCKOPT, &ServerIOsb, 0, 0,
                      SOL_SOCKET, SO_REUSEADDR,
                      &OptionEnabled, sizeof(OptionEnabled), 0, 0);
   if (Debug)
      fprintf (stdout, "sys$qiow() %%X%08.08X IOsb.Status %%X%08.08X\n",
               status, ServerIOsb.Status);
   if (VMSok (status) && VMSnok (ServerIOsb.Status))
      status = ServerIOsb.Status;
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "set socket option", __FILE__, __LINE__);
#endif

   /* bind the server socket to the host address and port */

   ServerSocketName.sin_family = AF_INET;
   ServerSocketName.sin_port = htons (ServerPort);
   ServerSocketName.sin_addr.s_addr = INADDR_ANY;

#ifdef IP_UCX
   if (Debug) fprintf (stdout, "sys$qiow() IO$_SETMODE\n");
   status = sys$qiow (0, ServerChannel, IO$_SETMODE, &ServerIOsb, 0, 0,
                      0, 0, &ServerSocketNameItem, QueueLength, 0, 0);
   if (Debug)
      fprintf (stdout, "sys$qiow() %%X%08.08X IOsb.Status %%X%08.08X\n",
               status, ServerIOsb.Status);
   if (VMSok (status) && VMSnok (ServerIOsb.Status))
      status = ServerIOsb.Status;
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "bind server socket", __FILE__, __LINE__);
#endif

#ifdef IP_MULTINET
   if (Debug) fprintf (stdout, "sys$qiow() IO$_BIND\n");
   status = sys$qiow (0, ServerChannel, IO$_BIND, &ServerIOsb, 0, 0,
                      &ServerSocketName, sizeof(ServerSocketName), 0, 0, 0, 0);
   if (Debug)
      fprintf (stdout, "sys$qiow() %%X%08.08X IOsb.Status %%X%08.08X\n",
               status, ServerIOsb.Status);
   if (VMSok (status) && VMSnok (ServerIOsb.Status))
      status = ServerIOsb.Status;
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "bind server socket", __FILE__, __LINE__);

   if (Debug) fprintf (stdout, "sys$qio() IO$_LISTEN\n");
   status = sys$qiow (0, ServerChannel, IO$_LISTEN, &ServerIOsb, 0, 0,
                      QueueLength, 0, 0, 0, 0, 0);
   if (Debug)
      fprintf (stdout, "sys$qiow() %%X%08.08X IOsb.Status %%X%08.08X\n",
               status, ServerIOsb.Status);
   if (VMSok (status) && VMSnok (ServerIOsb.Status))
      status = ServerIOsb.Status;
   if (VMSnok (status))
      ErrorExitVmsStatus (status, "listen server socket", __FILE__, __LINE__);
#endif

   /* turn off SYSPRV after binding to well-known port */
   if (VMSnok (status = sys$setprv (0, &SysPrvMask, 0, 0)))
      ErrorExitVmsStatus (status, "disabling SYSPRV", __FILE__, __LINE__);

   /************************************************/
   /* loop listening for and accepting connections */
   /************************************************/

   for (;;)
   {
      /* assign a channel to the internet template device */
      if (VMSnok (status = sys$assign (&InetDeviceDsc, &ClientChannel, 0, 0)))
         ErrorExitVmsStatus (status, "assign client channel",
                             __FILE__, __LINE__);

#ifdef IP_UCX
      if (Debug) fprintf (stdout, "sys$qio() IO$_ACCESS | IO$M_ACCEPT\n");
      status = sys$qio (0, ServerChannel, IO$_ACCESS | IO$M_ACCEPT,
                        &ServerIOsb, &NetConnectRequestAst, 0,
                        0, 0, &ClientSocketNameItem, &ClientChannel,
                        &LingerOnCloseOption, 0);

#endif

#ifdef IP_MULTINET
      if (Debug) fprintf (stdout, "sys$qio() IO$_ACCEPT\n");
      status = sys$qio (0, ClientChannel, IO$_ACCEPT,
                        &ServerIOsb, &NetConnectRequestAst, 0,
                        &ClientSocketName, sizeof(ClientSocketName),
                        ServerChannel, 0, 0, 0);
#endif

      if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
      if (VMSnok (status)) 
         ErrorExitVmsStatus (status, "accept", __FILE__, __LINE__);

      /* reenable user-mode ASTs (needed on subsequent times through loop) */
      sys$setast (1);

      /*********************************/
      /* wait for a connection request */
      /*********************************/

      sys$hiber ();

      /*******************************/
      /* connection request received */
      /*******************************/

      /* NOTE: user-mode ASTs are now disabled by NetConnectRequestAst() */

      if (Debug)
         fprintf (stdout, "sys$qio() IOsb.Status %%X%08.08X\n",
                  ServerIOsb.Status);
      
      if (ServerIOsb.Status == SS$_ABORT && 
          ControlExitRequested || ControlRestartRequested)
      {
         /* server exiting or restarting, reenable user-mode ASTs */
         sys$setast (1);
         /* continue to hibernate until the server exits */
         for (;;) sys$hiber ();
      }

      if (VMSnok (ServerIOsb.Status)) 
         ErrorExitVmsStatus (ServerIOsb.Status, "accept", __FILE__, __LINE__);

      Accounting.ConnectCount++;

      ClientInternetAddress32bit = ClientSocketName.sin_addr.S_un.S_addr;

      strcpy (ClientInternetAddress, inet_ntoa (ClientSocketName.sin_addr));
      if (Debug) fprintf (stdout, "|%s|\n", ClientInternetAddress);
                                                                        
      if (Config.DnsLookup)
         NetHostLookup ();
      else
         strcpy (ClientHostName, ClientInternetAddress);
      if (Debug) fprintf (stdout, "|%s|\n", ClientHostName);

      if (CurrentConnectCount >= MaxConcurrentConnections)
      {
         Accounting.ConnectTooBusyCount++;
         Accounting.ResponseStatusCodeCountArray[5]++;
         NetTooBusy (ClientChannel);
         NetDummyRead (ClientChannel);
      }
      else
      if (ConfigAcceptClientHostName (ClientHostName))
      {
         Accounting.ConnectAcceptedCount++;
         /* allocate zeroed dynamic memory for the connection thread */
         if ((rqptr = calloc (1, sizeof(struct RequestStruct))) == NULL)
         {
            /* memory allocation failed */
            Accounting.ConnectCallocFailCount++;
            Accounting.ResponseStatusCodeCountArray[5]++;
            NetRequestCallocFailed (ClientChannel);
            NetDummyRead (ClientChannel);
         }
         else
         {
            /***************************/
            /* process HTTP connection */
            /***************************/

            /* once the thread structure has been created we have a client! */
            Accounting.ConnectCurrent = ++CurrentConnectCount;
            if (CurrentConnectCount > Accounting.ConnectPeak)
               Accounting.ConnectPeak = CurrentConnectCount;

            if (VMSnok (HttpdConnection (rqptr)))
               ConcludeProcessing (rqptr);
         }
      }
      else
      {
         Accounting.ConnectRejectedCount++;
         Accounting.ResponseStatusCodeCountArray[4]++;
         NetAccessDenied (ClientChannel);
         NetDummyRead (ClientChannel);
      }
   }
}

/*****************************************************************************/
/*
Disable user-mode ASTs at the start of connection acceptance processing and 
wake the hibernating process.
*/ 

NetConnectRequestAst ()

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetConnectRequestAst()\n");

   sys$setast (0);
   sys$wake (0, 0);
}

/*****************************************************************************/
/*
Some browsers report a network error (e.g. Windows Netscape Navigator 2.n) if
the server does not read anything from them before returning an error message.
This routine and it's AST ensure a read is done before reporting an error
earlier that reading the request from it.  Needless-to-say, no status checking
is performed and the data just discarded (hope it works consistently :^)
*/ 

NetDummyRead (unsigned short Channel)

{
   static char  DummyBuffer [1024];
   static struct AnIOsb  DummyIOsb;

   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetDummyRead() Channel: %d\n", Channel);

#ifdef IP_UCX
   status = sys$qio (0, Channel,
                     IO$_READVBLK, &DummyIOsb,
                     &NetDummyReadAst, Channel,
                     DummyBuffer, sizeof(DummyBuffer), 0, 0, 0, 0);
#endif

#ifdef IP_MULTINET
   status = sys$qio (0, Channel,
                     IO$_RECEIVE, &DummyIOsb,
                     &NetDummyReadAst, Channel,
                     DummyBuffer, sizeof(DummyBuffer), 0, 0, 0, 0);
#endif

   if (Debug) fprintf (stdout, "sys$qio() %%X%08.08X\n", status);
}

/*****************************************************************************/
/*
See commentary in NetDummyRead().
*/ 

NetDummyReadAst (unsigned short Channel)

{
   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetDummyReadAst() Channel: %d\n", Channel);

   NetShutdownSocket (Channel);
}

/*****************************************************************************/
/*
Perform the UCX get host name using QIO ACP functionality.  This can execute 
at AST-delivery level.  However, the only way I found of doing this within 
MultiNet was to use the gethostbyaddr() function, which seems to require USER-
mode ASTs to be enabled (not altogether surprising seeing it probably wasn't 
designed to be non-blocking).  Hence this kludge.  Turning on user-mode ASTs 
around gethostbyaddr() can slow down new connection processing significantly 
as AST-delivery is pre-emptive.  NetHostLookup() is not AST-driven and hence 
can introduce some granularity, particularly if DNS is involved in resolving 
the host name.
*/ 

NetHostLookup () 

{
#ifdef IP_UCX
   static unsigned char ControlSubFunction [4] =
      { INETACP_FUNC$C_GETHOSTBYADDR, INETACP$C_TRANS, 0, 0 };
   static struct dsc$descriptor AddressDsc =
      { 4, DSC$K_CLASS_S, DSC$K_DTYPE_T, &ClientSocketName.sin_addr };
   static struct dsc$descriptor ControlSubFunctionDsc =
      { 4, DSC$K_CLASS_S, DSC$K_DTYPE_T, &ControlSubFunction };
   static struct dsc$descriptor HostNameDsc =
      { sizeof(ClientHostName)-1, DSC$K_CLASS_S, DSC$K_DTYPE_T,
        &ClientHostName };
#endif

   int  status;
   unsigned short  Length;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetHostLookup()\n");

#ifdef IP_UCX
   Length = 0;
   if (Debug) fprintf (stdout, "sys$qio() IO$_ACPCONTROL\n");
   status = sys$qiow (0, ServerChannel, IO$_ACPCONTROL, &ServerIOsb, 0, 0,
                      &ControlSubFunctionDsc,
                      &AddressDsc, &Length, &HostNameDsc, 0, 0);
   if (Debug)
      fprintf (stdout, "sys$qio() %%X%08.08X IOsb.Status %%X%08.08X\n",
               status, ServerIOsb.Status);
   if (Length)
      ClientHostName[Length] = '\0';
   else
      strcpy (ClientHostName, ClientInternetAddress);
#endif

#ifdef IP_MULTINET
   /* reenable user-mode ASTs, to allow gethostbyaddr() to work */
   sys$setast (1);

   if (Debug) fprintf (stdout, "gethostbyaddr()\n");
   HostEntryPtr = gethostbyaddr (&ClientSocketName.sin_addr,
                                 sizeof(ClientSocketName.sin_addr),
                                 AF_INET);

   /* re-disable user-mode ASTs */
   sys$setast (0);

   if (Debug) fprintf (stdout, "HostEntryPtr: %d\n", HostEntryPtr);
   if (HostEntryPtr == NULL)
      strcpy (ClientHostName, ClientInternetAddress);
   else
      strcpy (ClientHostName, HostEntryPtr->h_name);
#endif
}

/****************************************************************************/
/*
*/

NetShutdownSocket (unsigned short Channel)

{
   int  status;
   struct  AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetShutdownSocket() %d\n", Channel);

#ifdef IP_UCX
      status = sys$qiow (0, Channel, IO$_DEACCESS | IO$M_SHUTDOWN, &IOsb, 0, 0,
                         0, 0, 0, 0, 0, 0);
#endif

#ifdef IP_MULTINET
      status = sys$qiow (0, Channel, IO$_SHUTDOWN, &IOsb, 0, 0,
                         0, 0, 0, 0, 0, 0);
#endif

   if (Debug)
      fprintf (stdout, "sys$qiow() %%X%08.08X IOsb.Status %%X%08.08X\n",
               status, ServerIOsb.Status);

   status = sys$dassgn (Channel);
   if (Debug) fprintf (stdout, "sys$dassgn() %%X%08.08X\n", status);
}

/****************************************************************************/
/*
Stop the server channel from receiving incoming requests.
*/

NetShutdownIncomingSocket (unsigned short Channel)

{
   int  status;
   struct  AnIOsb  IOsb;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetShutdownIncomingSocket() %d\n", Channel);

#ifdef IP_UCX
      status = sys$qiow (0, Channel, IO$_DEACCESS, &IOsb, 0, 0,
                         0, 0, 0, 0, 0, 0);
#endif

#ifdef IP_MULTINET
      status = sys$qiow (0, Channel, IO$_SHUTDOWN, &IOsb, 0, 0,
                         0, 0, 0, 0, 0, 0);
#endif

   if (Debug)
      fprintf (stdout, "sys$qiow() %%X%08.08X IOsb.Status %%X%08.08X\n",
               status, ServerIOsb.Status);
}

/*****************************************************************************/
/*
Unlike NetWriteQio() this function writes to a client channel, not using the 
full request data structure.  This is for writing to the client before any of 
the request data structures are created (e.g. when the server denies access).
This is a blocking function, but shouldn't introduce much granularity as it is
only used for short messages before a client thread is set up.
*/ 

int NetWrite
(
unsigned short Channel,
char *DataPtr,
int DataLength
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetWrite()\n");

#ifdef IP_UCX
   status = sys$qiow (0, Channel, IO$_WRITEVBLK, 0, 0, 0,
                      DataPtr, DataLength, 0, 0, 0, 0);
#endif

#ifdef IP_MULTINET
   status = sys$qiow (0, Channel, IO$_SEND, 0, 0, 0,
                      DataPtr, DataLength, 0, 0, 0, 0);
#endif

   return (status);
}

/*****************************************************************************/
/*
*/ 

NetResponseHeaderAst (struct RequestStruct *rqptr)

{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetResponseHeaderAst()\n");

   if (VMSnok (status =
       QioNetWrite (rqptr,
                    rqptr->ResponseHeaderAstAddress,
                    rqptr->ResponseHeaderDataPtr,
                    rqptr->ResponseHeaderDataLength)))
   {
      if (VMSnok (status =
          sys$dclast (rqptr->ResponseHeaderAstAddress, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
   }
}

/*****************************************************************************/
/*
Write 'DataLength' bytes located at 'DataPtr' via the network to the client.  
If 'AstCompletionAddress' zero then use sys$qiow(), waiting for completion.
If an AST completion address is supplied the use sys$qio().  If the HTTP
method is "HEAD" only allow bytes in the header to be sent, absorb any other,
explicitly calling the AST completion routine as necessary.
*/ 

int QioNetWrite
(
struct RequestStruct *rqptr,
void *AstCompletionAddress,
char *DataPtr,
int DataLength
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug)
      fprintf (stdout, "QioNetWrite() %d %d %d %d bytes\n",
               rqptr, AstCompletionAddress, DataPtr, DataLength);
   /** if (Debug) fprintf (stdout, "|%*s|\n", DataLength, DataPtr); **/

   status = SS$_NORMAL;

   if (!rqptr->BytesTx && rqptr->ResponseHeaderPtr)
   {
      /******************************/
      /* send response header first */
      /******************************/

      if (Debug)
         fprintf (stdout, "flushing rqptr->ResponseHeaderPtr\n|%s|\n",
                  rqptr->ResponseHeaderPtr);
      if (rqptr->HttpMethod == HTTP_METHOD_HEAD)
         rqptr->ResponseHeaderNewlineCount = 2;
      if (!rqptr->ResponseHeaderLength)
         rqptr->ResponseHeaderLength = strlen(rqptr->ResponseHeaderPtr);

      if (AstCompletionAddress)
      {
         if (DataLength)
         {
            rqptr->ResponseHeaderAstAddress = AstCompletionAddress;
            rqptr->ResponseHeaderDataPtr = DataPtr;
            rqptr->ResponseHeaderDataLength = DataLength;
            AstCompletionAddress = &NetResponseHeaderAst;
         }

#ifdef IP_UCX
         status = sys$qio (0, rqptr->ClientChannel,
                           IO$_WRITEVBLK, &rqptr->NetWriteIOsb,
                           AstCompletionAddress, rqptr,
                           rqptr->ResponseHeaderPtr,
                           rqptr->ResponseHeaderLength, 0, 0, 0, 0);
#endif

#ifdef IP_MULTINET
         status = sys$qio (0, rqptr->ClientChannel,
                           IO$_SEND, &rqptr->NetWriteIOsb,
                           AstCompletionAddress, rqptr,
                           rqptr->ResponseHeaderPtr,
                           rqptr->ResponseHeaderLength, 0, 0, 0, 0);
#endif
         if (Debug) fprintf (stdout, "sys$qio() %%%08.08X\n", status);

         if (VMSok (status))
            rqptr->BytesTx += rqptr->ResponseHeaderLength;
         else
         /* if resource wait enabled the only quota not waited for is ASTLM */
         if (status == SS$_EXQUOTA)
            ErrorExitVmsStatus (status, "sys$qio()", __FILE__, __LINE__);

         /* return, the AST function will sys$qiow() the data */
         return (status);
      }
      else
      {
#ifdef IP_UCX
         status = sys$qiow (0, rqptr->ClientChannel,
                            IO$_WRITEVBLK, &rqptr->NetWriteIOsb, 0, 0,
                            rqptr->ResponseHeaderPtr,
                            rqptr->ResponseHeaderLength, 0, 0, 0, 0);
#endif

#ifdef IP_MULTINET
         status = sys$qiow (0, rqptr->ClientChannel,
                            IO$_SEND, &rqptr->NetWriteIOsb, 0, 0,
                            rqptr->ResponseHeaderPtr,
                            rqptr->ResponseHeaderLength, 0, 0, 0, 0);
#endif
         if (Debug) fprintf (stdout, "sys$qio() %%%08.08X\n", status);

         if (VMSok (status))
            rqptr->BytesTx += rqptr->ResponseHeaderLength;
         else
         /* if resource wait enabled the only quota not waited for is ASTLM */
         if (status == SS$_EXQUOTA)
            ErrorExitVmsStatus (status, "sys$qio()", __FILE__, __LINE__);

         /* continue on to sys$qiow() the data */
      }
   }

   if (!DataLength)
   {
      if (!AstCompletionAddress) return (status);

      if (VMSnok (status =
          sys$dclast (AstCompletionAddress, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);

      return (status);
   }

   if (rqptr->HttpMethod == HTTP_METHOD_HEAD)
   {
      /*****************************/
      /* send only response header */
      /*****************************/

      if (Debug) fprintf (stdout, "method HEAD\n");
      if (rqptr->ResponseHeaderNewlineCount <= 1)
      {
         /* local storage */
         register int  cnt, nlcnt;
         register char  *cptr;

         nlcnt = rqptr->ResponseHeaderNewlineCount;
         if (Debug) fprintf (stdout, "nlcnt: %d\n", nlcnt);
         cptr = DataPtr;
         cnt = DataLength;
         while (cnt--)
         {
            if (*cptr == '\n' && ++nlcnt == 2)
            {
               /* two successive end-of-lines, therefore end of header */
               if (Debug) fprintf (stdout, "end-of-header\n");
               cptr++;
               break;
            }
            else
            if (*cptr != '\r' && *cptr != '\n')
               nlcnt = 0;
            cptr++;
         }
         if ((rqptr->ResponseHeaderNewlineCount = nlcnt) == 2)
         {
            if (Debug) fprintf (stdout, "DataLength: %d\n", cptr - DataPtr);
            if (DataLength = cptr - DataPtr)
            {
               /* adjust data length to include only the HTTP header */
               DataLength = cptr - DataPtr;
            }
            else
            {
               /* no data in HTTP header left at all */
               rqptr->NetWriteIOsb.Status = STS$K_SUCCESS;
               rqptr->NetWriteIOsb.Count = 0;
               if (AstCompletionAddress)
               {
                  if (VMSnok (status =
                      sys$dclast (AstCompletionAddress, rqptr, 0)))
                     ErrorExitVmsStatus (status, "sys$dclast()",
                                         __FILE__, __LINE__);
               }
               return;
            }
         }
      }
      else
      {
         /* HTTP header has been completely sent, absorb anything else */
         rqptr->NetWriteIOsb.Status = STS$K_SUCCESS;
         rqptr->NetWriteIOsb.Count = 0;
         if (AstCompletionAddress)
         {
            if (VMSnok (status =
                sys$dclast (AstCompletionAddress, rqptr, 0)))
               ErrorExitVmsStatus (status, "sys$dclast()",
                                   __FILE__, __LINE__);
         }
         return;
      }
   }

   /*************/
   /* send data */
   /*************/

   if (AstCompletionAddress)
   {
#ifdef IP_UCX
      status = sys$qio (0, rqptr->ClientChannel,
                        IO$_WRITEVBLK, &rqptr->NetWriteIOsb,
                        AstCompletionAddress, rqptr,
                        DataPtr, DataLength, 0, 0, 0, 0);
#endif

#ifdef IP_MULTINET
      status = sys$qio (0, rqptr->ClientChannel,
                        IO$_SEND, &rqptr->NetWriteIOsb,
                        AstCompletionAddress, rqptr,
                        DataPtr, DataLength, 0, 0, 0, 0);
#endif
   }
   else
   {
#ifdef IP_UCX
      status = sys$qiow (0, rqptr->ClientChannel,
                         IO$_WRITEVBLK, &rqptr->NetWriteIOsb, 0, 0,
                         DataPtr, DataLength, 0, 0, 0, 0);
#endif

#ifdef IP_MULTINET
      status = sys$qiow (0, rqptr->ClientChannel,
                         IO$_SEND, &rqptr->NetWriteIOsb, 0, 0,
                         DataPtr, DataLength, 0, 0, 0, 0);
#endif
   }
   if (Debug) fprintf (stdout, "sys$qio() %%%08.08X\n", status);

   if (VMSok (status))
      rqptr->BytesTx += DataLength;
   else
   /* if resource wait enabled the only quota not waited for is ASTLM */
   if (status == SS$_EXQUOTA)
      ErrorExitVmsStatus (status, "sys$qio()", __FILE__, __LINE__);

   return (status);
}

/*****************************************************************************/
/*
Queue up a read from the client over the network. If 'AstCompletionAddress' 
is zero then no I/O completion AST routine is called.  If it is non-zero then 
the function pointed to by the parameter is called when the network write 
completes.
*/ 

int QioNetRead
(
struct RequestStruct *rqptr,
void *AstCompletionAddress,
char *DataBuffer,
int SizeofDataBuffer
)
{
   int  status;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "QioNetRead()\n");

#ifdef IP_UCX
   status = sys$qio (0, rqptr->ClientChannel,
                     IO$_READVBLK, &rqptr->NetReadIOsb,
                     AstCompletionAddress, rqptr,
                     DataBuffer, SizeofDataBuffer, 0, 0, 0, 0);
#endif

#ifdef IP_MULTINET
   status = sys$qio (0, rqptr->ClientChannel,
                     IO$_RECEIVE, &rqptr->NetReadIOsb,
                     AstCompletionAddress, rqptr,
                     DataBuffer, SizeofDataBuffer, 0, 0, 0, 0);
#endif

   /* with resource wait enabled the only quota not waited for is ASTLM */
   if (status == SS$_EXQUOTA)
      ErrorExitVmsStatus (status, "sys$qio()", __FILE__, __LINE__);

   return (status);
}

/*****************************************************************************/
/*
This (moderately complex) function buffers multiple 'records' (usually lines 
of text) before writing them as one, larger 'record' via the network to the 
client.  It will escape HTML-forbidden characters if the appropriate flag is 
set. 

This function "double" buffers data.  This scheme allows the buffer to
"overflow" and be written via the network to the client without overwriting
that buffered data  when buffering the new data, and with processing remaining
correct for an AST-driven environment.  This single NetBufferOutput() call per
AST routine is very  critical to the integrity of the functionality!  This
requires an AST to be supplied with every call supplying data to be output
(i.e. every call EXCEPT an initial allocation call), AND that no more output be
buffered until that AST is delivered (i.e. only one NetBufferOutput() per
AST-context function.

Uses the thread structure's 'OutputBufferPtr', 'OutputBufferCurrentPtr' and
'OutputBufferRemaining' storage.  Call with 'OutputPtr' set to NULL to
explicitly flush the buffer at any time.  For uniformity of behaviour, an AST 
function MAY be declared with any call, if it's not employed in an actual 
network write, its explicitly declared.  An AST always sets the network write 
IOsb (it's fudged as SS$_NORMAL normal with declared ASTs to simulate an
actual AST-driven write).

The UCX TCP/IP driver (4.n at least!) appears to be able to support more than
65k in a single QIO!
*/ 

int NetBufferOutput
(
struct RequestStruct *rqptr,
void *AstFunctionPtr,
char *OutputPtr,
int OutputLength
)
{
   int  status,
        CurrentLength;
   unsigned char  *TempDoubleBufferPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug)
       fprintf (stdout, "NetBufferOutput() %d %d %d\n",
                OutputPtr, OutputLength, AstFunctionPtr);

   if (rqptr->OutputBufferPtr == NULL)
   {
      /**********************************************/
      /* allocate heap memory for the output buffer */
      /**********************************************/

      if (Debug) fprintf (stdout, "allocate\n");

      /* allow two bytes for possible null terminator */
      if ((rqptr->OutputBufferPtr =
          HeapAlloc (rqptr, OutputBufferSize+1)) == NULL)
      {
         ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
         return (SS$_INSFMEM);
      }
      rqptr->OutputBufferCurrentPtr = rqptr->OutputBufferPtr;
      rqptr->OutputBufferRemaining =
         rqptr->OutputBufferSize = OutputBufferSize;

      if (OutputPtr == NULL) return (SS$_NORMAL);
   }

   if (!AstFunctionPtr)
      ErrorExitVmsStatus (status, ErrorSanityCheck, __FILE__, __LINE__);

   if (OutputPtr == NULL)
   {
      /*************************/
      /* flush buffer contents */
      /*************************/

      if (Debug)
         fprintf (stdout, "flush %d bytes\n|%s|\n",
                  rqptr->OutputBufferCurrentPtr - rqptr->OutputBufferPtr,
                  rqptr->OutputBufferPtr);

      if (rqptr->OutputBufferCurrentPtr > rqptr->OutputBufferPtr)
      {
         if (VMSnok (status =
             QioNetWrite (rqptr, AstFunctionPtr,
                          rqptr->OutputBufferPtr,
                          rqptr->OutputBufferCurrentPtr -
                             rqptr->OutputBufferPtr)))
            ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
      }
      else
      {
         /* no network write actually occured, declare an AST to do it */
         if (VMSnok (status = sys$dclast (AstFunctionPtr, rqptr, 0)))
            ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
         status = rqptr->NetWriteIOsb.Status = SS$_NORMAL;
      }

      rqptr->OutputBufferCurrentPtr = rqptr->OutputBufferPtr;
      rqptr->OutputBufferRemaining = OutputBufferSize;

      return (status);
   }

   if (Debug)
      fprintf (stdout, "[%d] %d\n|%*s|\n",
               rqptr->NetBufferOutputEscapeHtml,
               OutputLength, OutputLength, OutputPtr);

   if (rqptr->NetBufferOutputEscapeHtml)
   {
      /************************************/
      /* escape HTML-forbidden characters */
      /************************************/

      /* local storage */
      register int  bcnt, ocnt;
      register unsigned char  *bptr, *optr;

      if (Debug) fprintf (stdout, "escape-HTML\n");

      TempDoubleBufferPtr = NULL;
      optr = OutputPtr;
      ocnt = OutputLength;
      bptr = rqptr->OutputBufferCurrentPtr;
      bcnt = rqptr->OutputBufferRemaining;

      while (ocnt)
      {
         /* worst-case is having to escape an ampersand */
         if (bcnt < 5)
         {
            if (TempDoubleBufferPtr != NULL)
            {
               /*
                  If there should be so many HTML-escaped characters that
                  the output buffer needs flushing twice ... then don't,
                  its not designed to allow that.  Instead wriggle out of
                  it gracefully (as possible), returning an error message.
                  This should seldom happen, if at all!
               */
               ErrorVmsStatus (rqptr, RMS$_RTB, __FILE__, __LINE__);
               return (RMS$_RTB);
            }

            /* only provide the double-buffering when necessary */
            if (rqptr->OutputDoubleBufferPtr == NULL)
            {
               if (Debug) fprintf (stdout, "double-buffering\n");
               if ((rqptr->OutputDoubleBufferPtr =
                   HeapAlloc (rqptr, rqptr->OutputBufferSize+2))
                  == NULL)
               {
                  ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
                  return (SS$_INSFMEM);
               }
            }

            status = QioNetWrite (rqptr, AstFunctionPtr,
                                  rqptr->OutputBufferPtr,
                                  bptr - rqptr->OutputBufferPtr);

            TempDoubleBufferPtr = rqptr->OutputDoubleBufferPtr;
            rqptr->OutputDoubleBufferPtr = rqptr->OutputBufferPtr;
            rqptr->OutputBufferPtr = TempDoubleBufferPtr;

            bptr = rqptr->OutputBufferPtr;
            bcnt = OutputBufferSize;

            if (VMSnok (status))
            {
               ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
               return (status);
            }

            /* ensure any AST function is queued only once! */
            AstFunctionPtr = 0;
         }

         switch (*optr)
         {
            case '<' :
               memcpy (bptr, "&lt;", 4); bptr += 4; bcnt -= 4; optr++; break;
            case '>' :
               memcpy (bptr, "&gt;", 4); bptr += 4; bcnt -= 4; optr++; break;
            case '&' :
               memcpy (bptr, "&amp;", 5); bptr += 5; bcnt -= 5; optr++; break;
            default :
               *bptr++ = *optr++; bcnt--;
         }
         ocnt--;
      }

      rqptr->OutputBufferCurrentPtr = bptr;
      rqptr->OutputBufferRemaining = bcnt;
   }
   else
   {
      /*****************/
      /* buffer output */
      /*****************/

      if (OutputLength > rqptr->OutputBufferRemaining)
      {
         /************************************/
         /* insufficient space, flush buffer */
         /************************************/

         if (Debug)
            fprintf (stdout, "full-flush\n|%s|\n", rqptr->OutputBufferPtr);

         if (OutputLength > rqptr->OutputBufferSize)
         {
            /* record is actually too big to buffer! */
            ErrorVmsStatus (rqptr, RMS$_RTB, __FILE__, __LINE__);
            return (RMS$_RTB);
         }

         /* only provide the double-buffering when necessary */
         if (rqptr->OutputDoubleBufferPtr == NULL)
         {
            if (Debug) fprintf (stdout, "double-buffering\n");
            if ((rqptr->OutputDoubleBufferPtr =
                HeapAlloc (rqptr, rqptr->OutputBufferSize+2))
               == NULL)
            {
               ErrorHeapAlloc (rqptr, __FILE__, __LINE__);
               return (SS$_INSFMEM);
            }
         }

         status = QioNetWrite (rqptr, AstFunctionPtr,
                               rqptr->OutputBufferPtr,
                               rqptr->OutputBufferCurrentPtr -
                                  rqptr->OutputBufferPtr);

         TempDoubleBufferPtr = rqptr->OutputDoubleBufferPtr;
         rqptr->OutputDoubleBufferPtr = rqptr->OutputBufferPtr;
         rqptr->OutputBufferPtr = TempDoubleBufferPtr;

         rqptr->OutputBufferCurrentPtr = rqptr->OutputBufferPtr;
         rqptr->OutputBufferRemaining = OutputBufferSize;

         if (VMSnok (status))
         {
            ErrorVmsStatus (rqptr, status, __FILE__, __LINE__);
            return (status);
         }

         /* ensure any AST function is queued only once! */
         AstFunctionPtr = 0;
      }

      /* buffer new output */
      memcpy (rqptr->OutputBufferCurrentPtr, OutputPtr, OutputLength);
      rqptr->OutputBufferCurrentPtr += OutputLength;
      rqptr->OutputBufferRemaining -= OutputLength;
   }

   /******************/
   /* declare an AST */
   /******************/

   /* if a network write has not been done then explicitly declare the AST */
   if (AstFunctionPtr)
   {
      /* fudge the fact a network write has not actually been done! */
      rqptr->NetWriteIOsb.Status = SS$_NORMAL;
      if (VMSnok (status = sys$dclast (AstFunctionPtr, rqptr, 0)))
         ErrorExitVmsStatus (status, "sys$dclast()", __FILE__, __LINE__);
   }

   return (SS$_NORMAL);
}

/****************************************************************************/
/*
*/

NetGetServerHostName ()

{
   register char  *cptr;

   struct in_addr  ServerAddress;
   int  status;
   struct hostent  *HostEntryPtr;

   /*********/
   /* begin */
   /*********/

   if (Debug) fprintf (stdout, "NetGetServerHostName()\n");


   /* get local (server) host details */
   if (gethostname (ServerHostName, sizeof(ServerHostName)) != 0)
      ErrorExitVmsStatus (0, strerror(errno), __FILE__, __LINE__);
   if (Debug) fprintf (stdout, "|%s|\n", ServerHostName);

   /* looks better if its all in lower case (host name is sometimes upper) */
   for (cptr = ServerHostName; *cptr; cptr++) *cptr = tolower(*cptr);
   ServerHostNameLength = cptr - ServerHostName;

   if ((HostEntryPtr = gethostbyname (ServerHostName)) == NULL)
      ErrorExitVmsStatus (0, strerror(errno), __FILE__, __LINE__);

   memcpy (&ServerAddress, HostEntryPtr->h_addr, 4);
   strcpy (ServerInternetAddress, inet_ntoa (ServerAddress));
   if (Debug) fprintf (stdout, "|%s|\n", ServerInternetAddress);
}

/*****************************************************************************/
/*
Report that the initial memory allocation failed.
*/
 
NetRequestCallocFailed (unsigned short Channel)
 
{
   static char  ErrorMessage [] =
"HTTP/1.0 500 Request calloc() failed.\r\n\
Content-Type: text/html\r\n\
\r\n\
<HTML>\n\
<HEAD>\n\
<TITLE>Error 500</TITLE>\n\
</HEAD>\n\
<BODY>\n\
Request calloc() failed.\n\
<P><I>(notify system administrator)</I>\n\
</BODY>\n\
</HTML>\n";

   NetWrite (Channel, ErrorMessage, sizeof(ErrorMessage)-1);
}

/*****************************************************************************/
/*
Report that client host name has been denied access to the server.
*/
 
NetAccessDenied (unsigned short Channel)
 
{
   static char  ErrorMessage [] =
"HTTP/1.0 403 Server access denied.\r\n\
Content-Type: text/html\r\n\
\r\n\
<HTML>\n\
<HEAD>\n\
<TITLE>Error 403</TITLE>\n\
</HEAD>\n\
<BODY>\n\
Server access denied.\n\
</BODY>\n\
</HTML>\n";

   NetWrite (Channel, ErrorMessage, sizeof(ErrorMessage)-1);
}

/*****************************************************************************/
/*
Report that there is already the limit of client connections.
*/
 
NetTooBusy (unsigned short Channel)
 
{
   static char  ErrorMessage [] =
"HTTP/1.0 502 Server too busy.\r\n\
Content-Type: text/html\r\n\
\r\n\
<HTML>\n\
<HEAD>\n\
<TITLE>Error 502</TITLE>\n\
</HEAD>\n\
<BODY>\n\
Server too busy right now ... try again shortly.\n\
</BODY>\n\
</HTML>\n";

   NetWrite (Channel, ErrorMessage, sizeof(ErrorMessage)-1);
}

/****************************************************************************/

